/*
 *
 *  MakeLib
 *  Copyright (c) 1997-2001 Max Kellermann (max.kellermann@epost.de)
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation;
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: FileEmbeddedManager.java,v 1.9 2001/10/11 21:32:01 max Exp $
 *
 */

package make.datamodel.embedded;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.ByteArrayInputStream;
import java.io.OutputStream;
import java.io.FileOutputStream;
import java.io.FileInputStream;
import java.util.Hashtable;
import org.w3c.dom.Node;
import org.w3c.dom.Document;
import org.xml.sax.SAXException;
import make.io.Streams;
import make.xml.DOMUtil;
import make.xml.DocumentType;
import make.xml.SchemaManager;

/**
 * EmbeddedManager implementation that saves all data
 * in single files in one directory.
 */
public class FileEmbeddedManager implements EmbeddedManager {
    /*properties*****************************/
    protected SchemaManager schemaManager;
    protected File path, file;
    protected String encoding;
    /*Constructor*****************************/
    /**
     * Creates a new instance.
     * @param _f the document file. 
     *           All other entries are written into the same path.
     * @param _fm the SchemaManager which is used to lookup
     *            tips to write the XML file.
     * @param _enc the encoding that is used to write XML files.
     */
    public FileEmbeddedManager(File _f, SchemaManager _fm, String _enc) 
	throws IOException {
	if (!_f.isAbsolute())
	    _f = new File(_f.getAbsolutePath());
	file = _f;
	schemaManager = _fm;
	path = new File(file.getParent());
	encoding = _enc;
    }
    public String toString() { return file.toString(); }
    /*EmbeddedManager******************************************/
    public String[] getNames() {
	String[] list = path.list();
	int count = 0;
	for (int z=0; z<list.length; z++) {
	    File f = new File(path, list[z]);
	    if (f.isFile())
		list[count++] = list[z];
	}
	String[] names = new String[count];
	System.arraycopy(list, 0, names, 0, count);
	return names;
    }
    public Hashtable getSizes() {
	Hashtable ht = new Hashtable();
	String[] list = path.list();
	for (int z=0; z<list.length; z++) {
	    File f = new File(path, list[z]);
	    if (f.isFile())
		ht.put(list[z], new Long(f.length()));
	}
	return ht;
    }
    public boolean contains(String name) {
	File f = new File(path, name);
	return f.exists() && f.isFile();
    }
    public void replaceEntry(String name, InputStream data)
	throws IOException {
	File f = new File(path, name);
	File temp = new File(f.getPath()+".tmp");
	if (data!=null) {
	    FileOutputStream fos = new FileOutputStream(temp);
	    Streams.copy(data, fos);
	    fos.close();
	}
	if (f.exists())
	    f.delete();
	temp.renameTo(f);
    }
    public InputStream getEntry(String name) throws IOException {
	File f = new File(path, name);
	if (!f.exists() || !f.isFile())
	    return null;
	return new FileInputStream(f);
    }
    public void replaceDocument(Node n) throws IOException {
	final Document doc = n.getOwnerDocument();
	DocumentType type = null;
	if (schemaManager!=null) {
	    String tn = schemaManager.getSchemaName(doc);
	    if (tn!=null)
		type = schemaManager.getSchema(tn);
	}
	byte[] data = DOMUtil.toByteArray(doc, type, encoding);
	replaceEntry(file.getName(), new ByteArrayInputStream(data));
    }
    public Node getDocument() throws IOException, SAXException {
	return DOMUtil.parse(getEntry(file.getName())).getDocumentElement();
    }
}
