/*
 *
 *  MakeLib
 *  Copyright (c) 1997-2001 Max Kellermann (max.kellermann@epost.de)
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation;
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: ZipEmbeddedManager.java,v 1.10 2001/10/11 21:32:40 max Exp $
 *
 */

package make.datamodel.embedded;

import java.util.Enumeration;
import java.util.Vector;
import java.util.Hashtable;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.ByteArrayInputStream;
import java.io.OutputStream;
import java.io.FileOutputStream;
import java.util.zip.ZipFile;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;
import org.w3c.dom.Node;
import org.w3c.dom.Document;
import org.xml.sax.SAXException;
import make.io.Streams;
import make.xml.DOMUtil;
import make.xml.DocumentType;
import make.xml.SchemaManager;

/**
 * EmbeddedManager implementation that saves all data
 * in a .ZIP file. 
 * The XML document is saved as document.xml.
 */
public class ZipEmbeddedManager implements EmbeddedManager {
    /*CONST***********************************/
    public static final String DOCUMENT_NAME = "document.xml";
    public static final String DUMMY_NAME = "_dummy";
    /*properties*****************************/
    protected SchemaManager schemaManager;
    protected File file;
    protected ZipFile images;
    protected String encoding;
    /*Constructor*****************************/
    /**
     * Creates a new instance.
     * @param _f the zip file.
     * @param _fm the SchemaManager which is used to lookup
     *            tips to write the XML file.
     * @param _enc the encoding that is used to write XML files.
     */
    public ZipEmbeddedManager(File _f, SchemaManager _fm, String _enc) 
	throws IOException {
	file = _f; schemaManager = _fm;
	encoding = _enc;
	open();
    }
    public String toString() { return file.getName(); }
    /*internal**********************************/
    /**
     * Opens the ZIP file or creates a new file if open fails.
     */
    protected void open() throws IOException {
	try {
	    images = new ZipFile(file);
	} catch (IOException e) {
	    FileOutputStream fos = new FileOutputStream(file);
	    ZipOutputStream zos = new ZipOutputStream(fos);
	    zos.putNextEntry(new ZipEntry(DUMMY_NAME));
	    zos.closeEntry();
	    zos.close();
	    images = new ZipFile(file);
	}
    }
    /**
     * Copies an entry from one zip file to another one.
     * @param zos the destination zip file
     * @param zf the source zip file
     * @param ze the zip entry to copy
     */
    protected static void copy(ZipOutputStream zos, ZipFile zf, ZipEntry ze) 
	throws IOException {
	zos.putNextEntry(new ZipEntry(ze.getName()));
	Streams.copy(zf.getInputStream(ze), zos);
	zos.closeEntry();
    }
    /**
     * Copies an entry from one zip file to another one.
     * @param zos the destination zip file
     * @param zf the source zip file
     * @param nc the name of the entry to copy
     */
    protected static int copy(ZipOutputStream zos, ZipFile zf, String nc) 
	throws IOException {
	int count = 0;
	for (Enumeration enum=zf.entries(); enum.hasMoreElements();) {
	    ZipEntry ze = (ZipEntry)enum.nextElement();
	    if (!ze.getName().equals(nc)) {
		copy(zos, zf, ze);
		count++;
	    }
	}
	return count;
    }
    /**
     * Searches for an entry in the objects zip file.
     * @param name the name of the entry to find
     */
    protected ZipEntry findEntry(String name) {
	if (images==null) return null;
	for (Enumeration enum=images.entries(); enum.hasMoreElements();) {
	    ZipEntry ze = (ZipEntry)enum.nextElement();
	    if (name.equals(ze.getName()))
		return ze;
	}
	return null;
    }
    /*EmbeddedManager******************************************/
    public String[] getNames() {
	Vector entries = new Vector();
	for (Enumeration enum=images.entries(); enum.hasMoreElements();) {
	    ZipEntry ze = (ZipEntry)enum.nextElement();
	    entries.addElement(ze.getName());
	}
	String[] names = new String[entries.size()];
	int z=0;
	for (Enumeration enum=entries.elements(); enum.hasMoreElements();)
	    names[z++] = (String)enum.nextElement();
	return names;
    }
    public Hashtable getSizes() {
	Hashtable ht = new Hashtable();
	for (Enumeration enum=images.entries(); enum.hasMoreElements();) {
	    ZipEntry ze = (ZipEntry)enum.nextElement();
	    ht.put(ze.getName(), new Long(ze.getSize()));
	}
	return ht;
    }
    public boolean contains(String name) {
	return findEntry(name)!=null;
    }
    public void replaceEntry(String name, InputStream data)
	throws IOException {
	File temp = new File(file.getPath()+".tmp");
	try {
	    temp.delete();
	    FileOutputStream fos = new FileOutputStream(temp);
	    ZipOutputStream zos = new ZipOutputStream(fos);
	    final int count = copy(zos, images, name);
	    if (name!=null && data!=null) {
		zos.putNextEntry(new ZipEntry(name));
		Streams.copy(data, zos);
		zos.closeEntry();
	    } else if (count==0) {
		// write a dummy entry because zip files must not be empty
		zos.putNextEntry(new ZipEntry(DUMMY_NAME));
		zos.closeEntry();
	    }
	    zos.finish(); zos.close();
	    fos.close();
	    images.close(); file.delete();
	    temp.renameTo(file);
	    open();
	} catch (IOException e) {
	    temp.delete();
	    throw e;
	}
    }
    public InputStream getEntry(String name) throws IOException {
	ZipEntry ze = findEntry(name);
	return ze==null ? null : images.getInputStream(ze);
    }
    public void replaceDocument(Node n) throws IOException {
	final Document doc = n.getOwnerDocument();
	DocumentType type = null;
	if (schemaManager!=null) {
	    String tn = schemaManager.getSchemaName(doc);
	    if (tn!=null)
		type = schemaManager.getSchema(tn);
	}
	byte[] data = DOMUtil.toByteArray(doc, encoding, type);
	replaceEntry(DOCUMENT_NAME, new ByteArrayInputStream(data));
    }
    public Node getDocument() throws IOException, SAXException {
	InputStream is = getEntry(DOCUMENT_NAME);
	if (is==null)
	    return null;
	return DOMUtil.parse(is).getDocumentElement();
    }
}
