/*
 *
 *  MakeLib
 *  Copyright (c) 1997-2001 Max Kellermann (max.kellermann@epost.de)
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation;
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: Streams.java,v 1.11 2001/10/11 22:35:57 max Exp $
 *
 */

package make.io;

import java.io.IOException;
import java.io.EOFException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.ByteArrayOutputStream;
import java.io.OutputStreamWriter;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.Writer;

/**
 * Streams contains many static functions that help you
 * handle streams of bytes and characters.
 */
public abstract class Streams {
    /**
     * A buffer that is used by copy
     */
    private static final byte[] buffer = new byte[8192];
    /**
     * A flag indicating whether the buffer property is in use
     * by a thread calling copy
     */
    private static boolean bufferInUse = false;
    /*Static******************************/
    /**
     * Writes all data that can be read from the InputStream
     * into the OutputStream object.
     * This function stops after the specified number of bytes.
     */
    public static void copy(InputStream is, OutputStream os, int maxLen)
	throws IOException {
	int ch;
	while (maxLen-->0 && (ch=is.read())!=-1)
	    os.write(ch);
    }
    /**
     * Writes all data that can be read from the InputStream
     * into the OutputStream object.
     */
    public static void copy(InputStream is, OutputStream os)
	throws IOException {
	if (bufferInUse) {
	    int ch;
	    while ((ch=is.read())!=-1)
		os.write(ch);
	} else 
	    try {
		bufferInUse = true;
		synchronized(buffer) {
		    for (;;) {
			int len = is.read(buffer);
			if (len==-1) break;
			os.write(buffer, 0, len);
		    }
		}
	    } catch (EOFException e) {
	    } finally {
		bufferInUse = false;
	    }
    }
    /**
     * Writes all data that can be read from the InputStream
     * into both OutputStream objects. It is similar to
     * DualOutputStream.
     */
    public static void copy(InputStream is, OutputStream os1, OutputStream os2)
	throws IOException {
	int ch;
	while ((ch=is.read())!=-1) {
	    os1.write(ch);
	    os2.write(ch);
	}
    }
    /**
     * Writes all data that can be read from the Reader
     * into the Writer object.
     */
    public static void copy(Reader is, Writer os)
	throws IOException {
	int ch;
	while ((ch=is.read())!=-1)
	    os.write(ch);
    }
    /**
     * This method reads from a stream into an array 
     * until the array is full. It is more safe than just calling
     * the read method of an InputStream.
     */
    public static void read(InputStream is, byte[] array)
	throws IOException {
	int pos = 0;
	while (pos<array.length)
	    pos += is.read(array, pos, array.length-pos);
    }
    /**
     * This method reads from a stream into an array 
     * until the array is full. It is more safe than just calling
     * the read method of a Reader.
     */
    public static void read(Reader rd, char[] array)
	throws IOException {
	int pos = 0;
	while (pos<array.length)
	    pos += rd.read(array, pos, array.length-pos);
    }
    /**
     * Converts a string into a byte array using the specified encoding.
     * @param data the string to be converted
     * @param encoding the encoding to be used
     */
    public static byte[] toByteArray(String data, String encoding) 
	throws IOException {
	ByteArrayOutputStream baos = new ByteArrayOutputStream(data.length());
	OutputStreamWriter osw = new OutputStreamWriter(baos, encoding);
	osw.write(data);
	osw.flush();
	return baos.toByteArray();
    }
    /**
     * Reads the whole stream and returns the content as an array.
     * Be careful when reading large files! 
     * Best regards from OutOfMemoryError ;)
     */
    public static byte[] streamToByteArray(InputStream is)
	throws IOException {
	ByteArrayOutputStream baos = new ByteArrayOutputStream(is.available());
	Streams.copy(is, baos);
	return baos.toByteArray();
    }
    /**
     * Converts the specified character array into a byte array.
     */
    public static byte[] charArrayToByteArray(char[] src, String encoding) 
	throws IOException {
	ByteArrayOutputStream baos = new ByteArrayOutputStream(src.length*2);
	Writer wr = new OutputStreamWriter(baos, encoding);
	for (int z=0; z<src.length; z++)
	    wr.write((int)src[z]);
	wr.close();
	return baos.toByteArray();
    }
    /**
     * Converts the specified character array into a byte array
     * using the default encoding.
     */
    public static byte[] charArrayToByteArray(char[] src) 
	throws IOException {
	return charArrayToByteArray(src, defaultEncoding);
    }
    /**
     * Reads the contents of the stream (at a maximum of maxLen bytes),
     * and converts the data into a byte array.
     */
    public static byte[] inputStreamToByteArray(InputStream is, int maxLen) 
	throws IOException {
	ByteArrayOutputStream baos = new ByteArrayOutputStream();
	copy(is, baos, maxLen);
	return baos.toByteArray();
    }
    /**
     * Appends all the data that can be read from a Reader
     * object to a StringBuffer.
     */
    public static void append(StringBuffer sb, Reader rd) 
	throws IOException {
	int ch;
	while ((ch=rd.read())!=-1)
	    sb.append((char)ch);
    }
    /*Encoding****************************************/
    public static final String defaultEncoding;
    static {
	String _de;
	try {
	    _de = System.getProperty("encoding.default", "ISO-8859-1");
	} catch (SecurityException e) {
	    _de = "ISO-8859-1";
	}
	defaultEncoding = _de;
    }
    public static Reader streamToReader(InputStream is) throws IOException {
	return new InputStreamReader(is, defaultEncoding);
    }
    public static Writer streamToWriter(OutputStream os) throws IOException {
	return new OutputStreamWriter(os, defaultEncoding);
    }
}
