/*
 *
 *  MakeLib
 *  Copyright (c) 1997-2001 Max Kellermann (max.kellermann@epost.de)
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation;
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: MIMEUtil.java,v 1.10 2001/12/13 14:56:02 max Exp $
 *
 */

package make.mime;

import java.util.Hashtable;
import java.util.StringTokenizer;
import java.io.IOException;
import java.io.Reader;
import java.io.InputStreamReader;

/**
 * MIMEUtil contains static helper functions to work with
 * MIME messages.
 */
public abstract class MIMEUtil {
    /**
     * Checks if the message is of the specified content type.
     * Parameters to the Content-Type header are ignored when comparing.
     */
    public static boolean isContentType(Message msg, String type) {
        String msgType = msg.getHeader("Content-Type");
        if (msgType==null)
            return false;
        int idxSemi = msgType.indexOf(';');
        if (idxSemi!=-1)
            msgType = msgType.substring(0, idxSemi);
        return msgType.equals(type);
    }
    /**
     * Checks if the message is of the specified media type.
     * Examples for media types are 'image', 'text'.
     */
    public static boolean isMediaType(Message msg, String type) {
        String msgType = msg.getHeader("Content-Type");
        if (msgType==null)
            return false;
        return msgType.startsWith(type+"/") ||
            msgType.equals(type);
    }
    /**
     * Decodes all parameters contained in a MIME header value.
     * Returns a Hashtable object where parameter name maps to
     * parameter value. The main value of the header is mapped
     * with the key "" (empty string).
     */
    public static Hashtable decodeHeaderValue(String s) {
        Hashtable ht = new Hashtable(2);
        StringTokenizer st = new StringTokenizer(s, ";");
        while (st.hasMoreTokens()) {
            String token = st.nextToken();
            int idx = token.indexOf('=');
            String name = idx==-1?"":token.substring(0, idx);
            String value = idx==-1?token:token.substring(idx+1);
            value = value.trim();
            if ((value.startsWith("\"") && value.endsWith("\"")) ||
                (value.startsWith("'") && value.endsWith("'")))
                value = value.substring(1, value.length()-1);
            ht.put(name.trim(), value);
        }
        return ht;
    }
    /**
     * Extracts a parameter from a MIME header value.
     * See decodeHeaderValue for details about parameters in MIME headers.
     */
    public static String getParameter(String value, String parameter) {
        if (value==null)
            return null;
        Hashtable ht = decodeHeaderValue(value);
        return (String)ht.get(parameter);
    }
    /**
     * Returns the reader associated with the content of the 
     * specified message object.
     * The header Content-Type is evaluated to find out the charset.
     */
    public static Reader getReader(Message msg) 
        throws IOException {
        String charset = getParameter(msg.getHeader("Content-Type"), "charset");
        if (charset==null)
            charset = "us-ascii";
        return new InputStreamReader(msg.getInputStream(), charset);
    }
    /**
     * Reads the content of the message and turns it into a String.
     */
    public static String readString(Message msg)
        throws IOException {
        Reader rd = getReader(msg);
        String lenString = msg.getHeader("Content-Length");
        long len = lenString==null ? 256 : Long.parseLong(lenString);
        char[] data = new char[(int)len];
        int charsRead = rd.read(data);
        if (charsRead<0)
            charsRead = 0;
        return new String(data, 0, charsRead);
    }
    /**
     * Returns the name of the message. The headers Content-Disposition
     * or Content-Id are used to find out the name.
     */
    public static String getName(Message msg) {
        String name;
        name = getParameter(msg.getHeader("Content-Disposition"), "name");
        if (name!=null) return name;
        name = msg.getHeader("Content-Id");
        if (name!=null) return name;
        return null;
    }
}
