/*
 *
 *  MakeLib
 *  Copyright (c) 1997-2001 Max Kellermann (max.kellermann@epost.de)
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation;
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: MultiPartDecoder.java,v 1.7 2001/12/13 15:10:34 max Exp $
 *
 */

package make.mime;

import java.io.IOException;
import java.io.OutputStream;
import java.io.InputStream;
import java.io.PushbackInputStream;
import java.io.ByteArrayOutputStream;
import java.io.ByteArrayInputStream;
import java.util.Hashtable;
import java.util.StringTokenizer;
import make.util.Debug;
import make.io.NullOutputStream;

public class MultiPartDecoder {
    /***properties*****************************/
    protected byte[] boundary;
    protected PushbackInputStream in;
    protected boolean moreParts = true;
    /***Constructor***************************/
    public MultiPartDecoder(Message msg) 
        throws IOException {
        // extract boundary from Content-Type header
        String bnd = MIMEUtil.getParameter
            (msg.getHeader("Content-Type"), "boundary");
        boundary = new byte[bnd.length()+3];
        boundary[0] = (byte)'\n';
        boundary[1] = (byte)'-';
        boundary[2] = (byte)'-';
        for (int z=3; z<boundary.length; z++)
            boundary[z] = (byte)bnd.charAt(z-3);
        // create stream which can rewing a whole boundary length
        in = new PushbackInputStream(msg.getInputStream(), boundary.length);
        // skip trash before first boundary
        writeData(new NullOutputStream(), 1024*1024);
    }
    /***Util**********************************/
    public static Hashtable decodeHeaderValue(String s) {
        Hashtable ht = new Hashtable(2);
        StringTokenizer st = new StringTokenizer(s, ";");
        while (st.hasMoreTokens()) {
            String token = st.nextToken();
            int idx = token.indexOf('=');
            String name = idx==-1?"":token.substring(0, idx);
            String value = idx==-1?token:token.substring(idx+1);
            ht.put(name.trim(), value.trim());
        }
        return ht;
    }
    /***Internal*******************************/
    protected static byte[] readLine(InputStream is) 
        throws IOException {
        ByteArrayOutputStream os = new ByteArrayOutputStream(160);
        int ch;
        while ((ch=is.read())!=-1 && ch!='\r' && ch!='\n') {
            os.write(ch);
        }
        if (ch=='\r') {
            ch = is.read();
            if (Debug.DEBUG && ch!='\n')
                System.err.println("WARNING: MultiPartDecoder detected \\r without \\n");
        }
        return os.toByteArray();
    }
    protected Hashtable readHeader() throws IOException {
        Hashtable ht = new Hashtable();
        for (;;) {
            byte[] lnb = readLine(in);
            if (lnb.length==0)
                break;
            String line = new String(lnb, "us-ascii");
            int idx = line.indexOf(':');
            if (idx==-1)
                break;
            String name = line.substring(0, idx).trim();
            String value = line.substring(idx+1).trim();
            ht.put(name, value);
        }
        return ht;
    }
    protected void writeData(OutputStream out, long maxLen) 
        throws IOException {
        if (Debug.DEBUG && !moreParts)
            throw new InternalError("no more parts");
        int ch;
        long len = 0;
        boolean wasCR = false;
        for (;;) {
            ch = in.read();
            if (ch==-1)
                throw new IOException("unexpected EOF");
            if (ch=='\r') {
                if (wasCR)
                    out.write(ch);
                else
                    wasCR = true;
                continue;
            } else if (ch==boundary[0]) {
                int boundaryMatch = 0;
                do {
                    boundaryMatch++;
                    ch = in.read();
                } while (boundaryMatch<boundary.length && ch==boundary[boundaryMatch]);
                if (boundaryMatch==boundary.length) {
                    if (ch=='-')
                        moreParts = false;
                    else
                        in.unread(ch);
                    // skip rest of this line
                    readLine(in);
                    return;
                }
                if (ch==-1)
                    throw new IOException("unexpected EOF");
                in.unread(ch);
                if (boundaryMatch>1)
                    in.unread(boundary, 1, boundaryMatch-1);
                if (wasCR) {
                    in.unread('\r');
                    wasCR = false;
                }
                ch = boundary[0];
                boundaryMatch = 0;
            } else if (wasCR) {
                out.write('\r');
                wasCR = false;
            }
            if (len>=maxLen) 
                throw new IOException("part is too large");
            out.write(ch);
            len++;
        }
    }
    protected byte[] getData(int maxLen) 
        throws IOException {
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        writeData(baos, maxLen);
        byte[] data = baos.toByteArray();
        baos.close();
        return data;
    }
    protected InputStream getInputStream(int maxLen) 
        throws IOException {
        return new ByteArrayInputStream(getData(maxLen));
    }
    /***MultiPartDecoder******************************/
    public boolean hasMoreParts() {
        return moreParts;
    }
    public Message nextPart() throws IOException {
        Hashtable header = readHeader();
        byte[] data = getData(32*1024*1024);
        return new ArrayMessage(data, header);
    }
}
