/*
 *
 *  MakeLib
 *  Copyright (c) 1997-2001 Max Kellermann (max.kellermann@epost.de)
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation;
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: Scheduler.java,v 1.12 2001/08/20 09:23:39 max Exp $
 *
 */

package make.scheduler;

import java.util.Date;
import java.util.Vector;
import java.util.Enumeration;

/**
 * Scheduler is a daemon thread that performs actions
 * at a specified time. 
 * It may be useful to create indices for a database, remove
 * temporary files or records etc. - you only have to write 
 * a Task for it.
 */
public class Scheduler extends Thread {
    /*properties**************************/
    protected Vector tasks = new Vector();
    /*Constructor***************************/
    /**
     * Creates a new Scheduler.
     * @param _name thread name (for debug purposes)
     * @param _daemon specifies whether this process is a daemon process or not.
     *                If your application only creates a Scheduler and then
     *                exits the main thread, you have to specify false here!
     *                Please look at java.lang.Thread for details.
     */
    public Scheduler(String _name, boolean _daemon) {
	if (_name!=null)
	    setName(_name);
	setDaemon(_daemon);
	start();
    }
    /**
     * Creates a new Scheduler with a daemonized thread.
     */
    public Scheduler() {
	this(null, true);
    }
    /**
     * Creates a new Scheduler with a daemonized thread.
     * @param _name thread name (for debug purposes)
     */
    public Scheduler(String _name) {
	this(_name, true);
    }
    /*Scheduler***************************/
    /**
     * Stops the scheduler and frees all resources.
     */
    public void dispose() {
	tasks.removeAllElements();
	tasks = null;
	interrupt();
    }
    /**
     * Checks if there are tasks left.
     */
    public boolean isEmpty() {
	return tasks.isEmpty();
    }
    /**
     * Returns an Enumeration of all Task objects.
     */
    public Enumeration getTasks() {
	return tasks.elements();
    }
    /**
     * Adds a new Task to the Scheduler.
     */
    public void addTask(Task _task) {
	tasks.addElement(_task);
    }
    /**
     * Removes a Task from the Scheduler.
     */
    public void removeTask(Task _task) {
	tasks.removeElement(_task);
    }
    /*Runnable*****************************/
    public void run() {
	while (true) {
	    final long now = System.currentTimeMillis();
	    if (tasks==null)
		break;
	    for (Enumeration enum=tasks.elements(); enum.hasMoreElements();) {
		Task t = (Task)enum.nextElement();
		Date next = t.getNextTime();
		if (next==null) {
		    removeTask(t);
		} else if (now>=next.getTime()) {
		    t.run();
		}
		if (tasks==null)
		    break;
	    }
	    try {
		Thread.sleep(2000);
	    } catch (InterruptedException e) {
		if (tasks==null)
		    break;
	    }
	}
    }
}
