/*
 *
 *  MakeLib
 *  Copyright (c) 1997-2001 Max Kellermann (max.kellermann@epost.de)
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation;
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: TableMap.java,v 1.1 2001/12/14 17:08:34 max Exp $
 *
 */

package make.sql;

import java.util.Map;
import java.util.Collection;
import java.util.Set;
import java.util.Iterator;
import java.sql.SQLException;
import java.sql.Connection;
import java.sql.Statement;
import java.sql.PreparedStatement;
import java.sql.ResultSet;

public class TableMap implements Map {
    /*properties************************/
    protected PreparedStatement queryStatement, countStatement,
        updateStatement, deleteStatement, insertStatement;

    /*Constructor***************************/
    public TableMap(Connection connection, String tableName) 
        throws SQLException {
        // check if table exists
        try {
            countStatement.executeQuery().close();
        } catch (SQLException e) {
            // probably not - create it
            try {
                Statement s = connection.createStatement();
                s.executeUpdate
                    ("CREATE TABLE "+tableName+" ("+
                     "  name CHAR(128) NOT NULL PRIMARY KEY,"+
                     "  value CHAR(255) NOT NULL"+
                     ");");
                s.close();
            } catch (SQLException e2) {
                // this failed, but be throw the original exception instead.
                // most probably, the connection is faulty and the first exception
                // was not about a missing table.
                throw e;
            }
        }

        // prepare some statements
        queryStatement = connection.prepareStatement
            ("SELECT value FROM "+tableName+" WHERE name=?");
        countStatement = connection.prepareStatement
            ("SELECT COUNT(*) FROM "+tableName);
        updateStatement = connection.prepareStatement
            ("UPDATE "+tableName+" SET value=? WHERE key=?");
        deleteStatement = connection.prepareStatement
            ("DELETE FROM "+tableName+" WHERE key=?");
        insertStatement = connection.prepareStatement
            ("INSERT INTO"+tableName+"(value,key) VALUES(?,?)");

    }
    protected void finalize() {
        try {
            if (queryStatement!=null)
                queryStatement.close();
            if (countStatement!=null)
                countStatement.close();
            if (updateStatement!=null)
                updateStatement.close();
            if (deleteStatement!=null)
                deleteStatement.close();
            if (insertStatement!=null)
                insertStatement.close();
        } catch (SQLException e) {
        }
    }

    /*Map*************************************/
    public void clear() {
        throw new UnsupportedOperationException();
    }
    public boolean containsKey(Object _key) {
        return get(_key)!=null;
    }
    public boolean containsValue(Object _value) {
        throw new UnsupportedOperationException();
    }
    public Set entrySet() {
        throw new UnsupportedOperationException();
    }
    public boolean equals(Object o) {
        return o==this;
    }
    public synchronized Object get(Object _key) {
        String key = (String)_key;
        ResultSet rs = null;
        try {
            synchronized(queryStatement) {
                queryStatement.setString(1, key);
                rs = queryStatement.executeQuery();
                if (!rs.next())
                    return null;
                return rs.getString(1);
            }
        } catch (SQLException e) {
            e.printStackTrace();
            return null;
        } finally {
            try {
                if (rs!=null)
                    rs.close();
            } catch (SQLException e) {
                e.printStackTrace();
            }
        }
    }
    public int hashCode() {
        return 0;
    }
    public boolean isEmpty() {
        return size()==0;
    }
    public Set keySet() {
        throw new UnsupportedOperationException();
    }
    public synchronized Object put(Object _key, Object _value) {
        if (_value==null)
            return remove(_key);
        Object oldValue = get(_key);
        String key = (String)_key;
        String value = (String)_value;
        try {
            PreparedStatement statement = oldValue==null
                ? insertStatement : updateStatement;
            synchronized(statement) {
                statement.setString(1, value);
                statement.setString(2, key);
                statement.executeUpdate();
            }
        } catch (SQLException e) {
            e.printStackTrace();
        }
        return oldValue;
    }
    public void putAll(Map t) {
        Iterator it = t.entrySet().iterator();
        while (it.hasNext()) {
            Map.Entry entry = (Map.Entry)it.next();
            put(entry.getKey(), entry.getValue());
        }
    }
    public synchronized Object remove(Object _key) {
        Object oldValue = get(_key);
        if (oldValue==null)
            return null;
        String key = (String)_key;
        try {
            synchronized(deleteStatement) {
                deleteStatement.setString(1, key);
                deleteStatement.executeUpdate();
            }
        } catch (SQLException e) {
            e.printStackTrace();
        }
        return oldValue;
    }
    public synchronized int size() {
        ResultSet rs = null;
        try {
            synchronized(countStatement) {
                rs = countStatement.executeQuery();
                if (!rs.next())
                    return 0;
                return rs.getInt(1);
            }
        } catch (SQLException e) {
            e.printStackTrace();
            return 0;
        } finally {
            try {
                if (rs!=null)
                    rs.close();
            } catch (SQLException e) {
                e.printStackTrace();
            }
        }
    }
    public Collection values() {
        throw new UnsupportedOperationException();
    }
}
