/*
 *
 *  MakeLib
 *  Copyright (c) 1997-2001 Max Kellermann (max.kellermann@epost.de)
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation;
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: ButtonManager.java,v 1.14 2001/11/19 14:23:28 max Exp $
 *
 */

package make.swing;

import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Collection;
import java.util.Iterator;
import java.awt.event.ActionListener;
import javax.swing.Action;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import make.util.Debug;

/**
 * A ButtonManager instance manages a pool of javax.swing.Action
 * and make.swing.GenericButton objects. When an Action and
 * a button have got the same action command, it automatically
 * connects them.
 *
 * This class is very strict:
 * Before the ButtonManager is destructed, you must remove
 * all buttons and actions. Otherwise, this class will print a debug
 * message to System.err. You must also call dispose.
 */
final public class ButtonManager {
    /*ActionLink************************************/
    final static class ActionLink implements PropertyChangeListener {
        /*properties*************************/
        protected Action action;
        protected Vector buttons = new Vector();
        /*Constructor**************************/
        public ActionLink(Action _a) { 
            action = _a;
            if (action!=null)
                action.addPropertyChangeListener(this);
        }
        public ActionLink(GenericButton _ab) { 
            buttons.addElement(_ab);
            _ab.setEnabled(false);
        }
        public void dispose() {
            if (action!=null) {
                if (Debug.DEBUG)
                    System.err.println("BM.dispose: action="+action+" name="+action.getValue(Action.NAME));
                uninstallAction();
            }
            if (Debug.DEBUG && !buttons.isEmpty()) {
                System.err.println("BM.dispose: buttons.size="+buttons.size());
                for (Enumeration enum=buttons.elements(); enum.hasMoreElements();)
                    System.err.println("  - bt="+enum.nextElement());
                buttons.removeAllElements();
            }
        }
        /*bind*********************************/
        protected void bind(GenericButton ab) {
            ab.addActionListener(action);
            ab.setEnabled(action.isEnabled());
            Boolean selected = (Boolean)action.getValue("selected");
            ab.setSelected(selected!=null && selected.booleanValue());
        }
        protected void unbind(GenericButton ab) {
            ab.removeActionListener(action);
            ab.setEnabled(false);
            ab.setSelected(false);
        }
        /*(un)install****************************/
        public void installAction(Action _a) {
            if (Debug.DEBUG && action!=null)
                System.err.println("ERROR: action is already installed: "+
                                   action.getValue(action.NAME));
            action = _a;
            for (Enumeration enum=buttons.elements(); enum.hasMoreElements();)
                bind((GenericButton)enum.nextElement());
            action.addPropertyChangeListener(this);
        }
        public void uninstallAction() {
            action.removePropertyChangeListener(this);
            for (Enumeration enum=buttons.elements(); enum.hasMoreElements();)
                unbind((GenericButton)enum.nextElement());
            action = null;
        }
        public void installButton(GenericButton ab) {
            buttons.addElement(ab);
            if (action!=null)
                bind(ab);
            else
                ab.setEnabled(false);
        }
        public void uninstallButton(GenericButton ab) {
            buttons.removeElement(ab);
            if (action!=null)
                unbind(ab);
            else
                ab.setEnabled(false);
        }
        /*Info*****************************************/
        public Action getAction() { return action; }
        /*PropertyChangeListener*************************/
        public void propertyChange(PropertyChangeEvent e) {
            String propertyName = e.getPropertyName();
            /*if (propertyName.equals(Action.NAME)) {
              if (item instanceof JMenuItem) {
              String text = (String)e.getNewValue();
              item.setText(text);
              }
              } else */
            if (propertyName.equals("enabled")) {
                boolean enabled = ((Boolean)e.getNewValue()).booleanValue();
                for (Enumeration enum=buttons.elements(); enum.hasMoreElements();)
                    ((GenericButton)enum.nextElement()).setEnabled(enabled);
            } else if (propertyName.equals("selected")) {
                boolean selected = ((Boolean)e.getNewValue()).booleanValue();
                for (Enumeration enum=buttons.elements(); enum.hasMoreElements();)
                    ((GenericButton)enum.nextElement()).setSelected(selected);
            }
	    
        }
        /*Debug****************************/
        public void dump(String prefix) {
            System.err.println(prefix+"ActionLink action="+action+" {");
            for (Enumeration enum=buttons.elements(); enum.hasMoreElements();) {
                System.err.println(prefix+"  button="+enum.nextElement().getClass());
            }
            System.err.println(prefix+"}");
        }
    }
    /*properties****************************/
    protected Hashtable actions = new Hashtable();
    /*Constructor*****************************/
    protected void finalize() throws Throwable {
        if (actions!=null) {
            System.err.println("WARNING: bm.dispose was not called for "+this);
            for (Enumeration enum=actions.elements(); enum.hasMoreElements();)
                ((ActionLink)enum.nextElement()).dispose();
        }
        super.finalize();
    }
    /**
     * Frees all data structures in this object.
     * You must call this before this object is garbage collected.
     */
    public void dispose() {
        for (Enumeration enum=actions.elements(); enum.hasMoreElements();)
            ((ActionLink)enum.nextElement()).dispose();
        actions = null;
    }
    /*Replace**********************************/
    /**
     * Installs a new action.
     */
    public void installAction(Action a) {
        String name = (String)a.getValue(a.NAME);
        synchronized(actions) {
            ActionLink al = (ActionLink)actions.get(name);
            if (al==null) {
                al = new ActionLink(a);
                actions.put(name, al);
            } else
                al.installAction(a);
        }
    }
    /**
     * Uninstalls an action.
     * Fatal error if the action was not registered.
     */
    public void uninstallAction(Action a) {
        String name = (String)a.getValue(a.NAME);
        ActionLink al = (ActionLink)actions.get(a.getValue(a.NAME));
        if (Debug.DEBUG) {
            if (al==null) 
                throw new InternalError("action "+a.getValue(a.NAME)+
                                        " not found");
            if (al.action!=a)
                throw new InternalError("action mismatch name="+name+
                                        " al.action="+al.action+" a="+a);
        }
        al.uninstallAction();
    }
    /**
     * Installs a bunch of new actions.
     */
    public void installActions(Enumeration enum) {
        if (enum!=null)
            while (enum.hasMoreElements())
                installAction((Action)enum.nextElement());
    }
    /**
     * Uninstalls a bunch of actions.
     */
    public void uninstallActions(Enumeration enum) {
        if (enum!=null)
            while (enum.hasMoreElements())
                uninstallAction((Action)enum.nextElement());
    }
    /**
     * Installs a bunch of new actions.
     */
    public void installActions(Collection c) {
        for (Iterator it=c.iterator(); it.hasNext();)
            installAction((Action)it.next());
    }
    /**
     * Uninstalls a bunch of new actions.
     */
    public void uninstallActions(Collection c) {
        for (Iterator it=c.iterator(); it.hasNext();)
            uninstallAction((Action)it.next());
    }
    /**
     * Installs a bunch of new actions.
     */
    public void installActions(Action[] a) {
        if (a!=null) 
            for (int z=0; z<a.length; z++)
                installAction(a[z]);
    }
    /**
     * Uninstalls a bunch of actions.
     */
    public void uninstallActions(Action[] a) {
        if (a!=null)
            for (int z=0; z<a.length; z++)
                uninstallAction(a[z]);
    }
    /**
     * Installs a new button.
     */
    public void installButton(GenericButton b) {
        String cmd = b.getActionCommand();
        synchronized(actions) {
            ActionLink al = (ActionLink)actions.get(cmd);
            if (al==null) {
                al = new ActionLink(b);
                actions.put(cmd, al);
            } else
                al.installButton(b);
        }
    }
    /**
     * Uninstalls a button.
     * Fatal error if the button was not registered.
     */
    public void uninstallButton(GenericButton b) {
        ActionLink al = (ActionLink)actions.get(b.getActionCommand());
        al.uninstallButton(b);
    }
    /**
     * Installs a bunch of new buttons.
     */
    public void installButtons(GenericButton[] b) {
        for (int z=0; z<b.length; z++)
            installButton(b[z]);
    }
    /**
     * Uninstalls a bunch of buttons.
     */
    public void uninstallButtons(GenericButton[] b) {
        for (int z=0; z<b.length; z++)
            uninstallButton(b[z]);
    }
    /**
     * Installs a bunch of new buttons.
     */
    public void installButtons(Enumeration enum) {
        while (enum.hasMoreElements())
            installButton((GenericButton)enum.nextElement());
    }
    /**
     * Uninstalls a bunch of buttons.
     */
    public void uninstallButtons(Enumeration enum) {
        while (enum.hasMoreElements())
            uninstallButton((GenericButton)enum.nextElement());
    }
    /**
     * Installs a bunch of new buttons.
     */
    public void installButtons(Collection c) {
        for (Iterator it=c.iterator(); it.hasNext();)
            installButton((GenericButton)it.next());
    }
    /**
     * Uninstalls a bunch of buttons.
     */
    public void uninstallButtons(Collection c) {
        for (Iterator it=c.iterator(); it.hasNext();)
            uninstallButton((GenericButton)it.next());
    }
    /*Info************************************/
    /**
     * Looks for an action with the specified command.
     * @param cmd the command searched for
     */
    public Action getAction(String cmd) {
        ActionLink al = (ActionLink)actions.get(cmd);
        return al==null?null:al.getAction();
    }
    /*Debug****************************/
    /**
     * Dumps the contents of this object to System.err
     * @param prefix the indent
     */
    public void dump(String prefix) {
        System.err.println(prefix+"ButtonManager {");
        for (Enumeration enum=actions.keys(); enum.hasMoreElements();) {
            String name = (String)enum.nextElement();
            ActionLink al = (ActionLink)actions.get(name);
            al.dump(prefix+name+"  ");
        }
        System.err.println(prefix+"}");
    }
}
