/*
 *
 *  MakeLib
 *  Copyright (c) 1997-2001 Max Kellermann (max.kellermann@epost.de)
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation;
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: StatusWindow.java,v 1.11 2001/08/20 09:23:39 max Exp $
 *
 */

package make.swing;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.BorderLayout;
import java.awt.Cursor;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import javax.swing.JPanel;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JButton;
import make.swing.progress.ProgressIndicator;

public class StatusWindow 
    implements Runnable, ActionListener,
	       ProgressIndicator {
    /*properties***************************/
    protected Thread thread = new Thread(this);
    protected JFrame frame;

    /**
     * A label that gets frequently updated to contain the current
     * value of the text variable.
     */
    protected JLabel label;
    protected Thread jobThread;
    /**
     * The text to be displayed by <code>label</code>. A client class
     * may update this by calling <code>setText()</code>.
     */
    protected String text;
    /**
     * The frame's title string.
     */
    protected String title;
    /**
     * Number of milliseconds to wait before updateFrame() gets called
     * the first time. This is the minimum time the initial status
     * message is visible.
     */
    protected int wait;
    
    /**
     * Indicates whether the StatusWindow has a cancel button.
     */
    protected boolean hasCancel;

    /*Constructor**************************/
    public StatusWindow(String _text, String _title, int _wait, 
			boolean _cancel, Thread _jobThread) {
	text = _text; title = _title; wait = _wait; 
	hasCancel = _cancel; jobThread = _jobThread;
	thread.setPriority(Thread.MIN_PRIORITY);
	thread.start();
    }
    public StatusWindow(String _text, String _title, int _wait, 
			boolean _cancel, Runnable _job) {
	text = _text; title = _title; wait = _wait; 
	hasCancel = _cancel; 
	jobThread = new Thread(_job);
	jobThread.setPriority(Thread.NORM_PRIORITY);
	jobThread.start();
	thread.setPriority(Thread.MIN_PRIORITY);
	thread.start();
    }
    public StatusWindow(String _text, String _title, int _wait, 
			boolean _cancel) {
	text = _text; title = _title; wait = _wait; 
	hasCancel = _cancel; 
	if (hasCancel)
	    jobThread = Thread.currentThread();
	thread.setPriority(Thread.MIN_PRIORITY);
	thread.start();
    }
    /*ProgressIndicator**********************/
    public void setText(String t) {
	text = t;
    }
    public void setProgressBounds(int min, int max) {
	// not supported
    }
    public void setProgressValue(int val) {
	// not supported
    }
    public synchronized void close() {
	jobThread = null;
	thread.interrupt();
    }
    /*Internal*****************************/
    /**
     * Creates the frame with the given title and adds a panel containing the
     * text (and a cancel button if desired).
     */
    protected Component createPanel() {
	JPanel panel = new JPanel(new BorderLayout(), false);
	panel.setBorder(javax.swing.BorderFactory.createEmptyBorder(10,10,10,10));
	label = new JLabel(text); text = null;
	panel.add("Center", label);
	if (hasCancel) {
	    JButton cancel = new JButton("Abbrechen");
	    cancel.addActionListener(this);
	    panel.add("East", cancel);
	}
	return panel;
    }
    protected void createFrame() {
	frame = new JFrame(title);
	frame.setDefaultCloseOperation(frame.DO_NOTHING_ON_CLOSE);
	frame.setResizable(false);
	frame.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
	frame.getContentPane().add("Center", createPanel());
	frame.pack();
    }
    /**
     * Sets the frame size to half screen width * half screen height and shows
     * it.
     */
    protected void positionFrame() {
	if (frame==null)
	    return;
	Dimension screen = frame.getToolkit().getScreenSize();
	Dimension size = frame.getSize();
	frame.setLocation((screen.width-size.width)/2, 
			  (screen.height-size.height)/2);
	frame.show();
    }
    /**
     * If the text is non-null (that means it has been set via the setText()-
     * method) the label is being updated to show the new text.
     */
    protected void updateFrame() {
	if (text!=null && label!=null) {
	    label.setText(text);
	    text = null;
	}
    }
    /*ActionListener***********************/
    /**
     * Stops the thread. Called when the cancel button is pressed.
     */
    public void actionPerformed(ActionEvent evt) {
	if (thread.isAlive())
	    thread.interrupt();
    }
    /*Runnable****************************/
    /**
     * Starts the job if it's non-null, waits <code>wait</code>
     * milliseconds, creates and positions the frame, and
     * checks every 250 milliseconds whether the label needs to
     * be updated with a new text.
     */
    public void run() {
	try {
	    if (wait>0) {
		try {
		    if (jobThread==null) {
			Thread.sleep(wait);
		    } else {
			jobThread.join(wait); 
			// wait at most 'wait' milliseconds for this thread to die
		    }
		} catch (InterruptedException e) {
		    return;
		}
	    }
	    
	    if (jobThread!=null && !jobThread.isAlive()) {
		// jobThread has died in the meantime, i.e. it
		// has finished its task. So we don't need to
		// show a status window.
		return;
	    }
	    createFrame();
	    positionFrame();
	    // wait...
	    while (jobThread==null || jobThread.isAlive()) {
		try {
		    if (jobThread==null)
			Thread.sleep(250);
		    else
			jobThread.join(250);
		} catch (InterruptedException exc) {
		    return;
		}
		// update the frame in case the jobThread
		// has called setText() in the meantime
		updateFrame();
	    }
	} finally {
	    if (jobThread!=null && jobThread.isAlive())
		jobThread.interrupt();
	    if (frame!=null)
		frame.dispose();
	}
    }
}
