/*
 *
 *  MakeLib
 *  Copyright (c) 1997-2001 Max Kellermann (max.kellermann@epost.de)
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation;
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: LocaleSelect.java,v 1.8 2002/01/04 09:33:58 max Exp $
 *
 */

package make.swing.il;

import java.util.Collection;
import java.util.Iterator;
import java.util.Set;
import java.util.List;
import java.util.LinkedList;
import java.util.Locale;
import java.util.ResourceBundle;
import java.text.ParseException;
import java.awt.Component;
import java.awt.BorderLayout;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import javax.swing.JOptionPane;
import javax.swing.JComboBox;
import javax.swing.JPanel;
import javax.swing.JButton;
import javax.swing.JList;
import javax.swing.ComboBoxModel;
import javax.swing.ListCellRenderer;
import javax.swing.event.ListDataListener;
import javax.swing.event.ListDataEvent;
import org.w3c.dom.Element;
import make.util.Debug;
import make.util.Resources;
import make.datamodel.slot.LocalizedSlotContainer;

public class LocaleSelect extends DefaultInputLink
    implements ComboBoxModel, ListCellRenderer, ActionListener {
    /*CONST***********************/
    protected static final Object NULL_LOCALE = Boolean.FALSE;
    protected static final Object OTHER_LOCALE = Boolean.TRUE;

    /*properties**************************/
    protected Locale userLocale;
    protected Collection listeners = new LinkedList();
    protected LocalizedSlotContainer slotContainer;
    protected InputLink inputLink;
    protected List locales = new LinkedList();
    protected JPanel panel = new JPanel(new BorderLayout(), false);
    protected JComboBox comboBox = new JComboBox(this);
    protected ListCellRenderer oldCellRenderer;
    protected boolean ignoreReload = false, firstReload = true, noApply = false;
    protected ResourceBundle resources = Resources.getResourceBundle
        (LocaleSelect.class.getClassLoader(), "make.swing.il.LocaleSelect");

    /*Constructor************************/
    public LocaleSelect(Element vd, LocalizedSlotContainer _sc,
                        InputLink _il, Locale locale) {
        super(vd, locale);
        userLocale = locale;
        slotContainer = _sc;
        inputLink = _il;
        comboBox.setEditable(false);
        oldCellRenderer = comboBox.getRenderer();
        comboBox.setRenderer(this);
        panel.add("Center", comboBox);
        JButton deleteButton = new JButton("X");
        deleteButton.addActionListener(this);
        panel.add("East", deleteButton);
    }

    /*LocaleSelect*******************************/
    protected Locale findMatchingLocale(Locale pattern) {
        if (pattern==null)
            return null;
        Set refLocales = slotContainer.getReferencedLocales();
        if (refLocales.contains(pattern))
            return pattern;
        String language = pattern.getLanguage();
        if (language.equals(""))
            return null;
        Iterator it = refLocales.iterator();
        while (it.hasNext()) {
            Locale locale = (Locale)it.next();
            if (locale==null)
                continue;
            if (locale.getLanguage().equals(language))
                return locale;
        }
        return null;
    }
    protected synchronized void setLocale(Locale locale) {
        try {
            // save old values with old locale
            if (!noApply)
                inputLink.apply();
            // change combobox selection
            if (locale!=null && !locales.contains(locale)) {
                // the locale is not yet in the list - add it
                locales.add(locale);
                int index = locales.indexOf(locale)+1;
                fireIntervalAdded(index, index);
            }
            slotContainer.setLocale(locale);
            fireContentsChanged(-1, -1);
            // load new values with new locale
            try {
                ignoreReload = true;
                inputLink.reload();
            } finally {
                ignoreReload = false;
            }
        } catch (ParseException e) {
            // FIXME: better reaction on ParseException please!
            Debug.report(e);
        }
    }

    /*Widget**************************/
    public Component getComponent() {
        return panel;
    }

    /*InputLink***********************/
    public void reload() {
        if (ignoreReload)
            return;
        synchronized(locales) {
            int oldSize = locales.size();
            Iterator it = slotContainer.getReferencedLocales().iterator();
            while (it.hasNext()) {
                Locale locale = (Locale)it.next();
                if (locale!=null && !locales.contains(locale))
                    locales.add(locale);
            }
            Locale locale = slotContainer.getLocale();
            if (locale!=null && !locales.contains(locale))
                locales.add(locale);
            int newSize = locales.size();
            if (oldSize!=newSize)
                fireIntervalAdded(oldSize+1, newSize+1);

            if (firstReload) {
                if (slotContainer.getLocale()==null) {
                    try {
                        noApply = true;
                        setLocale(findMatchingLocale(userLocale));
                    } finally {
                        noApply = false;
                    }
                }
                firstReload = false;
            }
        }
    }
    public void apply() throws ParseException {
    }
    public void checkInput() throws ParseException {
    }

    /*ComboBoxModel********************/
    public void setSelectedItem(Object anItem) {
        if (anItem==null || anItem==NULL_LOCALE || "".equals(anItem)) {
            setLocale(null);
        } else if (anItem==OTHER_LOCALE) {
            String localeString = JOptionPane.showInputDialog
                (resources.getString("prompt.localeCode"));
            if (localeString==null)
                return;
            setSelectedItem(localeString);
        } else if (anItem instanceof Locale) {
            setLocale((Locale)anItem);
        } else if (anItem instanceof String) {
            // String arrived... check what it is
            // is it already selected? -> ignore
            Locale oldSelection = slotContainer.getLocale();
            if (oldSelection!=null &&
                oldSelection.toString().equals(anItem))
                return;
            // is it a Locale which is already in the list?
            Iterator it = locales.iterator();
            while (it.hasNext()) {
                Locale locale = (Locale)it.next();
                if (locale.toString().equals(anItem)) {
                    setSelectedItem(locale);
                    return;
                }
            }
            // not in list yet -> create and add
            String localeString = (String)anItem;
            String language = "", country = "", variant="";
            int idx0 = localeString.indexOf('_');
            if (idx0>=0) {
                language = localeString.substring(0, idx0);
                int idx1 = localeString.indexOf('_', idx0+1);
                if (idx1>=0) {
                    country = localeString.substring(idx0+1, idx1);
                    variant = localeString.substring(idx1+1);
                } else {
                    country = localeString.substring(idx0+1);
                }
            } else {
                language = localeString;
            }
            setLocale(new Locale(language, country, variant));
        } else {
            System.err.println("not a Locale: "+anItem);
        }
    }
    public Object getSelectedItem() {
        if (slotContainer==null)
            return NULL_LOCALE;
        Locale locale = slotContainer.getLocale();
        if (locale==null)
            return NULL_LOCALE;
        return locale;
    }

    /*ListModel*************************/
    public int getSize() {
        return locales.size()+2;
    }
    public Object getElementAt(int index) {
        if (index<=0)
            return NULL_LOCALE;
        if (index>locales.size())
            return OTHER_LOCALE;
        return locales.get(index-1);
    }
    public void addListDataListener(ListDataListener l) {
        listeners.add(l);
    }
    public void removeListDataListener(ListDataListener l) {
        listeners.remove(l);
    }

    /*ListModel Fire***********************/
    protected void fireContentsChanged(int idx0, int idx1) {
        ListDataEvent evt = new ListDataEvent
            (this, ListDataEvent.CONTENTS_CHANGED, idx0, idx1);
        Iterator it = listeners.iterator();
        while (it.hasNext())
            ((ListDataListener)it.next()).contentsChanged(evt);
    }
    protected void fireIntervalAdded(int idx0, int idx1) {
        ListDataEvent evt = new ListDataEvent
            (this, ListDataEvent.INTERVAL_ADDED, idx0, idx1);
        Iterator it = listeners.iterator();
        while (it.hasNext())
            ((ListDataListener)it.next()).intervalAdded(evt);
    }
    protected void fireIntervalRemoved(int idx0, int idx1) {
        ListDataEvent evt = new ListDataEvent
            (this, ListDataEvent.INTERVAL_REMOVED, idx0, idx1);
        Iterator it = listeners.iterator();
        while (it.hasNext())
            ((ListDataListener)it.next()).intervalRemoved(evt);
    }

    /*ListCellRenderer***************************/
    public Component getListCellRendererComponent(JList list,
                                                  Object value,
                                                  int index,
                                                  boolean isSelected,
                                                  boolean cellHasFocus) {
        if (value==null || value==NULL_LOCALE)
            value = "["+resources.getString("label.defaultLocale")+"]";
        else if (value==OTHER_LOCALE)
            value = "["+resources.getString("label.otherLocale")+"]";
        else
            value = value.toString();

        return oldCellRenderer.getListCellRendererComponent
            (list, value, index, isSelected, cellHasFocus);
    }

    /*ActionListener**************************/
    public void actionPerformed(ActionEvent evt) {
        Locale locale = slotContainer.getLocale();
        slotContainer.deleteLocale(locale);
        int index = locales.indexOf(locale);
        if (index>=0) {
            locales.remove(locale);
            fireIntervalRemoved(index+1, index+1);
        }
    }
}
