/*
 *
 *  MakeLib
 *  Copyright (c) 1997-2001 Max Kellermann (max.kellermann@epost.de)
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation;
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: AbstractInterfacePanel.java,v 1.1 2001/10/12 20:01:55 max Exp $
 *
 */

package make.swing.panel;

import java.awt.Component;
import java.awt.Container;
import java.awt.Frame;
import java.awt.Window;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
import javax.swing.JToolBar;
import javax.swing.JMenuBar;
import javax.swing.JComponent;
import make.util.Debug;
import make.swing.ButtonManager;
import make.swing.bt.FrameDescription;
import make.swing.bt.ButtonContainer;
import make.swing.bt.ButtonDescription;
import make.swing.bt.MenuDescription;

public abstract class AbstractInterfacePanel 
    implements InterfacePanel, ButtonContainer {
    /*Class***************************/
    protected class InstalledFrameDescription {
	/*properties********************/
	protected FrameDescription frameDescription;
	protected Vector buttons = new Vector();
	protected Vector components = new Vector();
	/*Constructor********************/
	public InstalledFrameDescription(FrameDescription fd) {
	    frameDescription = fd;
	    fd.installPosition(root);
	    JToolBar tb = fd.installToolBar(toolBar, buttons, components);
	    if (tb!=toolBar)
		setToolBar(tb);
	    JMenuBar mb = fd.installMenuBar(menuBar, buttons, components);
	    if (mb!=menuBar) {
		menuBar = mb;
		setMenuBar(menuBar);
	    }
	    Container bb = fd.installButtonBox(buttonBox, buttons, components);
	    if (bb!=buttonBox)
		setButtonBox(bb);
	    buttonManager.installButtons(buttons.elements());
	}
	public void uninstall() {
	    buttonManager.uninstallButtons(buttons.elements());
	    buttons = null;
	    for (Enumeration enum=components.elements(); enum.hasMoreElements();) {
		Component component = (Component)enum.nextElement();
		Container parent = component.getParent();
		if (Debug.DEBUG && parent==null)
		    throw new InternalError();
		parent.remove(component);
		if (parent.getComponentCount()==0 &&
		    component instanceof JComponent) {
		    JComponent jc = (JComponent)component;
		    Component menu = (Component)jc.getClientProperty(MenuDescription.MENU);
		    if (menu!=null) {
			Container mb = menu.getParent();
			mb.remove(menu);
			mb.validate();
		    }
		}
		parent.validate();
	    }
	    components = null;
	}
    }
    /*properties********************************/
    private JToolBar toolBar;
    private JMenuBar menuBar;
    private Frame frame;
    private Component root;
    private Container buttonBox;
    private Component component, south;
    private ButtonManager buttonManager = new ButtonManager();
    private PanelContent content;
    private Container container;
    private Hashtable frameDescriptions = new Hashtable();
    protected boolean packIt = true;
    /*Constructor**********************************/
    protected void initialize(Frame _frame, Component _root, Container _container) {
	if (Debug.DEBUG && (frame!=null || root!=null || container!=null))
	    throw new InternalError();
	frame = _frame; root = _root; container = _container;
    }
    protected void beforeShow() {
	if (packIt && root instanceof Window) {
	    ((Window)root).pack();
	    packIt = false;
	}
    }
    /*Internal***************************************/
    protected void setComponent(Component c) {
	if (component!=null)
	    container.remove(component);
	component = c;
	if (component!=null)
	    container.add("Center", component);
	container.validate();
    }
    public void setSouth(Component _s) {
	if (south!=null)
	    container.remove(south);
	south = _s;
	if (south!=null)
	    container.add("South", south);
	container.validate();
    }
    public PanelContent getContent() { return content; }
    protected void setToolBar(JToolBar tb) {
	if (toolBar!=null)
	    throw new InternalError();
	toolBar = tb;
	container.add("North", toolBar);
	container.validate();
    }
    protected void setButtonBox(Container _bb) {
	buttonBox = _bb;
	setSouth(buttonBox);
    }
    protected JToolBar getToolBar() { return toolBar; }
    protected JMenuBar getMenuBar() { return menuBar; }
    protected void updateFocus() {
	if (component instanceof JComponent) {
	    JComponent jcomp = (JComponent)component;
	    Component c = (Component)jcomp.getClientProperty("focusComponent");
	    if (c!=null)
		c.requestFocus();
	}
    }
    /*AbstractInterfacePanel******************************/
    protected abstract void setMenuBar(JMenuBar mb);
    /*InterfacePanel******************************/
    public void setContent(PanelContent c) {
	content = c;
	setComponent(content==null?null:content.getComponent());
	updateFocus();
    }
    public void addFrameDescription(FrameDescription fd) {
	frameDescriptions.put(fd, new InstalledFrameDescription(fd));
    }
    public void removeFrameDescription(FrameDescription fd) {
	InstalledFrameDescription ifd
	    = (InstalledFrameDescription)frameDescriptions.get(fd);
	ifd.uninstall();
	frameDescriptions.remove(fd);
    }
    public ButtonContainer getButtonContainer() {
	return this;
    }
    public ButtonManager getButtonManager() {
	return buttonManager;
    }
    public void setTitle(String s) {
    }
    public Frame getFrame() {
	return frame;
    }
    public void setVisible(boolean _visible) {
	if (_visible)
	    beforeShow();
	root.setVisible(_visible);
	Component parent = root.getParent();
	if (parent!=null)
	    parent.validate();
	updateFocus();
    }
    public void close() {
	if (content!=null) {
	    content.panelClosed();
	    setContent(null);
	}
	setSouth(null);
	buttonManager.dispose();
	buttonManager = null;
    }
    /*ButtonContainer********************************/
    public ButtonDescription getButton(String name) {
	for (Enumeration enum=frameDescriptions.elements(); enum.hasMoreElements();) {
	    InstalledFrameDescription ifd = (InstalledFrameDescription)enum.nextElement();
	    FrameDescription fd = ifd.frameDescription;
	    ButtonDescription bd = fd.getButton(name);
	    if (bd!=null)
		return bd;
	}
	return null;
    }
}
