/*
 *
 *  MakeLib
 *  Copyright (c) 1997-2001 Max Kellermann (max.kellermann@epost.de)
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation;
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: DragAndDropSupport.java,v 1.4 2001/10/30 17:05:25 max Exp $
 *
 */

package make.swing.text;

import java.awt.Cursor;
import java.awt.Point;
import java.awt.event.MouseListener;
import java.awt.event.MouseEvent;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.DataFlavor;
import java.awt.dnd.DnDConstants;
import java.awt.dnd.DropTarget;
import java.awt.dnd.DropTargetListener;
import java.awt.dnd.DropTargetEvent;
import java.awt.dnd.DropTargetDragEvent;
import java.awt.dnd.DropTargetDropEvent;
import java.awt.dnd.DragSource;
import java.awt.dnd.DragSourceListener;
import java.awt.dnd.DragSourceEvent;
import java.awt.dnd.DragSourceDragEvent;
import java.awt.dnd.DragSourceDropEvent;
import java.awt.dnd.DragGestureRecognizer;
import java.awt.dnd.MouseDragGestureRecognizer;
import java.awt.dnd.DragGestureListener;
import java.awt.dnd.DragGestureEvent;
import javax.swing.text.BadLocationException;
import make.util.Debug;

public class DragAndDropSupport 
    implements DropTargetListener,
               DragGestureListener,
               DragSourceListener,
               MouseListener {

    /*properties*************************/
    protected DefaultEditorKit editorKit;
    protected TextComponent textComponent;
    protected DropTarget dropTarget;
    protected DragSource dragSource = new DragSource();
    protected DragGestureRecognizer dragGestureRecognizer;
    protected boolean mousePressedOnSelection = false;

    /*Constructor*************************/
    public DragAndDropSupport(DefaultEditorKit _ek) {
        editorKit = _ek;
    }
    public void install(TextComponent c) {
        mousePressedOnSelection = false;
        textComponent = c;
        c.addMouseListener(this);
        dropTarget = new DropTarget(textComponent, this);
        dragGestureRecognizer = textComponent.getToolkit()
            .createDragGestureRecognizer
            (MouseDragGestureRecognizer.class, dragSource,
             textComponent, DnDConstants.ACTION_COPY, this);
    }
    public void uninstall(TextComponent c) {
        if (Debug.DEBUG && textComponent!=c)
            throw new InternalError();
        c.removeMouseListener(this);
        textComponent = null;
        mousePressedOnSelection = false;
    }

    /*Internal*****************************/
    protected boolean inSelection(Point pt) {
        Caret caret = textComponent.getCaret();
        if (caret==null || !caret.isSelectionVisible())
            return false;
        final int dot = caret.getDot();
        final int mark = caret.getMark();
        if (dot==mark || dot<0 || mark<0)
            return false;
        final int startPos = Math.min(dot, mark);
        final int endPos = Math.max(dot, mark);
        final int position = textComponent.getUI().viewToModel
            (textComponent, pt);
        return position>=startPos && position<=endPos;
    }
    protected Transferable copySelection(Point pt) {
        Caret caret = textComponent.getCaret();
        if (caret==null || !caret.isSelectionVisible())
            return null;
        final int dot = caret.getDot();
        final int mark = caret.getMark();
        if (dot==mark || dot<0 || mark<0)
            return null;
        final int startPos = Math.min(dot, mark);
        final int endPos = Math.max(dot, mark);
        final int position = textComponent.getUI().viewToModel
            (textComponent, pt);
        if (position<startPos || position>endPos)
            return null;

        Document document = textComponent.getDocument();
        if (document==null)
            return null;

        try {
            return editorKit.copy(document, startPos, endPos);
        } catch (BadLocationException e) {
            Debug.report(e);
            return null;
        }
    }

    /*DropTargetListener****************************/
    public void dragEnter(DropTargetDragEvent evt) {
        if ((evt.getDropAction()&DnDConstants.ACTION_COPY_OR_MOVE) == 0) {
            evt.rejectDrag();
            return;
        }
    }
    public void dragExit(DropTargetEvent dte) {}
    public void dragOver(DropTargetDragEvent evt) {}
    public void drop(DropTargetDropEvent evt) {
        if (DefaultEditorKit.DUMP_CLIPBOARD) {
            Transferable transferable = evt.getTransferable();
            System.err.println("drop: transferable = "+transferable);
            DataFlavor[] dataFlavors = transferable.getTransferDataFlavors();
            for (int z=0; z<dataFlavors.length; z++) {
                System.err.println("dataFlavors["+z+"] = "+dataFlavors[z]);
                System.err.println("    mimeType = "+dataFlavors[z].getMimeType());
                System.err.println("    class = "+dataFlavors[z].getRepresentationClass());
            }
        }

        if ((evt.getDropAction()&DnDConstants.ACTION_COPY_OR_MOVE) == 0) {
            evt.rejectDrop();
            return;
        }

        if (textComponent==null) {
            evt.rejectDrop();
            return;
        }

        Document document = textComponent.getDocument();
        if (document==null) {
            evt.rejectDrop();
            return;
        }

        int pos = textComponent.getUI().viewToModel(textComponent, evt.getLocation());
        if (pos==-1) {
            evt.rejectDrop();
            return;
        }

        try {
            textComponent.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
            evt.acceptDrop(DnDConstants.ACTION_COPY);

            Caret caret = textComponent.getCaret();
            if (caret!=null)
                caret.setDot(pos);

            boolean success = editorKit.paste(document, pos, evt.getTransferable());

            evt.dropComplete(success);
        } catch (Exception e) {
            e.printStackTrace();
            evt.dropComplete(false);
        } finally {
            textComponent.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
        }
    }
    public void dropActionChanged(DropTargetDragEvent evt) {}

    /*DragGestureListener***************************/
    public void dragGestureRecognized(DragGestureEvent evt) {
        if (!mousePressedOnSelection)
            return;
        Transferable transferable = copySelection(evt.getDragOrigin());
        if (transferable==null)
            return;
        evt.startDrag(null, transferable, this);
    }

    /*DragSourceListener*************************/
    public void dragEnter(DragSourceDragEvent evt) {}
    public void dragOver(DragSourceDragEvent evt) {}
    public void dropActionChanged(DragSourceDragEvent evt) {}
    public void dragExit(DragSourceEvent evt) {}
    public void dragDropEnd(DragSourceDropEvent evt) {}

    /*MouseListener********************************/
    public void mouseClicked(MouseEvent evt) {}
    public void mouseEntered(MouseEvent evt) {}
    public void mouseExited(MouseEvent evt) {}
    public void mousePressed(MouseEvent evt) {
        /* It is not nice to disable drag'n'drop when no mouse button
         was pressed over a selection, as there may be other gestures
         than dragging the mouse to start a drag'n'drop operation.
         Just for now, this hack should be ok .. */
        mousePressedOnSelection = inSelection(evt.getPoint());
    }
    public void mouseReleased(MouseEvent evt) {}

}
