/*
 *
 *  MakeLib
 *  Copyright (c) 1997-2001 Max Kellermann (max.kellermann@epost.de)
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation;
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: ReplaceDialog.java,v 1.7 2001/10/12 19:59:09 max Exp $
 *
 */

package make.swing.text;

import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.Component;
import java.awt.Container;
import java.awt.event.ActionEvent;
import java.util.ResourceBundle;
import javax.swing.Action;
import javax.swing.AbstractAction;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.text.Segment;
import javax.swing.text.BadLocationException;
import make.util.Debug;
import make.swing.bt.FrameDescription;
import make.swing.bt.ButtonDescription;
import make.swing.panel.InterfacePanel;
import make.swing.panel.PanelContent;
import make.swing.field.Field;
import make.swing.field.TextField;
import make.swing.field.CheckBoxField;

public class ReplaceDialog implements PanelContent {
    /*static************************/
    protected static ResourceBundle resources = ResourceBundle.getBundle("make.swing.text.ReplaceDialog");
    protected static final ButtonDescription[] buttons = {
	new ButtonDescription("search", resources.getString("search.label")),
	new ButtonDescription("replace", resources.getString("replace.label")),
	new ButtonDescription("replaceAll", resources.getString("replaceAll.label")),
	new ButtonDescription("close-frame", resources.getString("close.label")),
    };
    protected static final FrameDescription frameDescription;
    public static final GridBagConstraints labelConstraints = new GridBagConstraints();
    public static final GridBagConstraints fieldConstraints = new GridBagConstraints();
    static {
	frameDescription = new FrameDescription(resources.getString("title"));
	frameDescription.setButtonBox(buttons);
	labelConstraints.fill = labelConstraints.HORIZONTAL;
	fieldConstraints.fill = fieldConstraints.BOTH;
	fieldConstraints.gridwidth = fieldConstraints.REMAINDER;
	fieldConstraints.weightx = 1.0;
    }
    /*properties************************/
    protected JPanel panel = new JPanel(new GridBagLayout(), false);
    protected TextField searchTextField = new TextField();
    protected TextField replaceTextField = new TextField();
    protected Field ignoreCaseField = new CheckBoxField
	(resources.getString("ignoreCase.label"), 
	 Boolean.TRUE, Boolean.FALSE);
    protected TextComponent textcomp;
    protected Document doc;
    protected Caret caret;
    protected InterfacePanel interfacePanel;
    protected boolean replace = false;
    protected boolean replacePossible = false;
    protected boolean replaceAll = false;
    protected boolean closed = false;
    protected int lastFoundPos = -1;
    /*Constructor*************************/
    public ReplaceDialog() {
	panel.setBorder(BorderFactory.createEmptyBorder(10,10,10,10));
	searchTextField.setActionCommand("searchEnter");
	replaceTextField.setActionCommand("replaceEnter");
	addField(panel,resources.getString("searchFor.label") , searchTextField);
	addField(panel,resources.getString("replaceWith.label") , replaceTextField);
	panel.add(ignoreCaseField.getComponent(), fieldConstraints);
    }

    public static void addField(Container parent, String _label, Field field) {
	JLabel label = new JLabel(_label);
	parent.add(label, labelConstraints);
	parent.add(field.getComponent(), fieldConstraints);
	label.setLabelFor(field.getComponent());
    }

    public void showDialog(InterfacePanel _ip, TextComponent _textcomp){
	interfacePanel = _ip;
	textcomp = _textcomp;
	doc = _textcomp.getDocument(); 
	caret = _textcomp.getCaret();
	interfacePanel.addFrameDescription(frameDescription);
	searchTextField.install(interfacePanel.getButtonManager());
	replaceTextField.install(interfacePanel.getButtonManager());
	ignoreCaseField.install(interfacePanel.getButtonManager());
	interfacePanel.getButtonManager().installActions(actions);
	interfacePanel.setContent(this);
	interfacePanel.setVisible(true);
    }
    
    protected void search(String searchText, boolean ignoreCase) {
        replace=false;
	if (!searchText(searchText, ignoreCase)) //endReached;
	    panel.getToolkit().beep(); 
    }
    protected void replace(String searchText, String replaceText, 
			   boolean ignoreCase) {
	replace=true;
	if(!replaceText(searchText, replaceText, ignoreCase)) //endReached;
	    panel.getToolkit().beep();
	if(!replacePossible)
	    replaceText(searchText, replaceText, ignoreCase);
    }
    protected void replaceAll(String searchText, String replaceText, 
			      boolean ignoreCase) {
	replace=true;
	caret.setDot(0);
	lastFoundPos=-1;
	replacePossible=true;
	replaceAll=true;
	doc.beginUpdate();
	try{
	    while (replaceText(searchText, replaceText, ignoreCase)){
	    }
	}
	finally {
	    doc.endUpdate(); 
	}
	replaceAll=false;
    }
  
    protected boolean searchText(String searchText, boolean ignoreCase) {
	int searchLength = searchText.length();
	int foundAt=0;
	if(searchLength>0){
	    if(lastFoundPos==caret.getDot())
		caret.setDot(caret.getDot()+1);
	    foundAt=findText(caret.getDot(), searchText, ignoreCase);
	    if(foundAt>=0){
		caret.setDot(foundAt+searchLength);
		caret.moveDot(foundAt);
		textcomp.requestFocus();
		return true;
	    }
	}
	caret.setDot(0);
	replacePossible=false;
	return false;
    }
	
    protected boolean replaceText(String searchText, String replaceText, 
				  boolean ignoreCase) {
	int searchLength = searchText.length();
	int foundAt=0;
	if(searchLength>0){
	    foundAt=findText(caret.getDot(), searchText, ignoreCase);
	    if(foundAt>=0){
		caret.setDot(foundAt+searchLength);
		caret.moveDot(foundAt);
		if (replacePossible){
		    textcomp.replaceSelection(replaceText,null);
		    caret.setDot(lastFoundPos+replaceText.length());
		    if(!replaceAll)
			replacePossible=false;
		} else {
		    replacePossible=true;
		}
		textcomp.requestFocus();
		return true;
	    }
	}
	caret.setDot(0);
	replacePossible=false;
	return false;
    }
    
    public int findTextnew(int fromPos, String searchText) {
	Segment seg = new Segment();
	try{
	    long milstart=System.currentTimeMillis();
	    doc.getText(0,doc.getRootElement().getEndOffset(),seg);
	    int pos = seg.toString().indexOf(searchText,fromPos);
	    lastFoundPos=pos;
	    return pos; 
	}
	catch (BadLocationException e){
	    panel.getToolkit().beep(); 
	    e.printStackTrace();
	}
	return -1;
    }
    public int findText(int fromPos, String searchText, boolean ignoreCase) {
	Segment seg = new Segment();
	int searchLength = searchText.length();
	int caretPos = caret.getDot();
	try{
	    long milstart=System.currentTimeMillis();
	    doc.getText(0,doc.getRootElement().getEndOffset(),seg);
	    if (ignoreCase) {
		String testString=seg.toString();
		String test;
		while(doc.getRootElement().getEndOffset()>caretPos+searchLength){
		    test=testString.substring(caretPos,caretPos+searchLength);
		    if(test.equalsIgnoreCase(searchText)){
			lastFoundPos=caretPos;
			return caretPos;
		    }
		    caretPos++;
		}
	    } else { // hier kann schnellere Systemroutine benutzt werden.
		int pos = seg.toString().indexOf(searchText,fromPos);
		lastFoundPos=pos;
		return pos; 
	    }
	}
	catch (BadLocationException e){
	    panel.getToolkit().beep(); 
	    e.printStackTrace();
	}
	return -1; //falls nix gefunden;
    }
    public int findTextok(int fromPos, String searchText, boolean ignoreCase){
	Segment seg = new Segment();
	int searchLength = searchText.length();
	int caretPos = caret.getDot();
	try{
	    long milstart=System.currentTimeMillis();
	    while(doc.getRootElement().getEndOffset()>caretPos+searchLength){
		doc.getText(caretPos,searchLength,seg);
		if( ignoreCase) {
		    if(seg.toString().equalsIgnoreCase(searchText)){
			lastFoundPos=caretPos;
			return caretPos;
		    }
		} else{
		    if(seg.toString().equals(searchText)){
			lastFoundPos=caretPos;
			return caretPos;
		    }
		}
		caretPos++;
	    }
	}
	catch (BadLocationException e){
	    panel.getToolkit().beep(); 
	    e.printStackTrace();
	}
	return -1; //falls nix gefunden;
    }

    /*PanelContent*****************************/
    public boolean panelClosing() { return true; }
    public void panelClosed() {
	doc = null; caret = null;
	interfacePanel.removeFrameDescription(frameDescription);
	searchTextField.uninstall(interfacePanel.getButtonManager());
	replaceTextField.uninstall(interfacePanel.getButtonManager());
	ignoreCaseField.uninstall(interfacePanel.getButtonManager());
	interfacePanel.getButtonManager().uninstallActions(actions);
	interfacePanel = null;
	textcomp.requestFocus();
    }
    public Component getComponent() { return panel; }
    /*Actions***************************/
    protected Action[] actions = new Action[]{
	new ReplaceAction(),
	new ReplaceAllAction(),
	new SearchAction(),
	new SearchEnterAction(),
	new ReplaceEnterAction(),
    };
    abstract class MyAction extends AbstractAction {
	MyAction(String ac) { super(ac); }
	protected abstract void actionPerformed(String searchText, 
						String replaceText,
						boolean ignoreCase);
        public void actionPerformed(ActionEvent evt) {
	    if (doc==null || caret==null) return;
	    try {
		String searchText = (String)searchTextField.getValue();
		String replaceText = (String)replaceTextField.getValue();
		boolean ignoreCase = ((Boolean)ignoreCaseField.getValue())
		    .booleanValue();
		actionPerformed(searchText, replaceText, ignoreCase);
	    } catch (Exception e) {
		Debug.report(e);
	    }
	}
    }	
    class ReplaceAction extends MyAction {
	ReplaceAction() { super("replace"); }
        public void actionPerformed(String searchText, 
				    String replaceText,
				    boolean ignoreCase) {
	    replace(searchText, replaceText, ignoreCase);
	}
    }
    class ReplaceAllAction extends MyAction {
	ReplaceAllAction() { super("replaceAll"); }
        public void actionPerformed(String searchText, 
				    String replaceText,
				    boolean ignoreCase) {
	    replaceAll(searchText, replaceText, ignoreCase);
	}
    }
    class SearchAction extends MyAction {
	SearchAction() { super("search"); }
        public void actionPerformed(String searchText, 
				    String replaceText,
				    boolean ignoreCase) {
	    search(searchText, ignoreCase);
	}
    }
    class SearchEnterAction extends MyAction {
	SearchEnterAction() { super("searchEnter"); }
        public void actionPerformed(String searchText, 
				    String replaceText,
				    boolean ignoreCase) {
	    replaceTextField.getComponent().requestFocus();
	}
    }
    class ReplaceEnterAction extends MyAction {
	ReplaceEnterAction() { super("replaceEnter"); }
        public void actionPerformed(String searchText, 
				    String replaceText,
				    boolean ignoreCase) {
	    if (replaceText==null) {
		search(searchText, ignoreCase);
	    } else {
		replace(searchText, replaceText, ignoreCase);
	    }
	}
    }
}
