/*
 *
 *  MakeLib
 *  Copyright (c) 1997-2001 Max Kellermann (max.kellermann@epost.de)
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation;
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: TextComponent.java,v 1.7 2001/10/30 20:58:18 max Exp $
 *
 */

package make.swing.text;

import java.util.ResourceBundle;
import java.awt.Color;
import java.awt.Insets;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Dimension;
import java.awt.AWTEvent;
import java.awt.event.KeyEvent;
import java.awt.event.ActionEvent;
import java.awt.event.MouseEvent;
import javax.swing.JComponent;
import javax.swing.Scrollable;
import javax.swing.Action;
import javax.swing.AbstractAction;
import javax.swing.KeyStroke;
import javax.swing.SwingConstants;
import javax.swing.JViewport;
import javax.swing.ToolTipManager;
import javax.swing.text.Position;
import javax.swing.text.BadLocationException;
import make.swing.ButtonManager;
import make.swing.text.view.Highlighter;

public abstract class TextComponent extends JComponent 
    implements Scrollable {
    /*CONST*******************************/
    public static final ResourceBundle resources
        = ResourceBundle.getBundle("make.swing.text.Messages");
    public static final String BUTTON_MANAGER = "buttonManager";
    public static final String DOCUMENT = "document";
    public static final String EDITOR_KIT = "editorKit";
    public static final String CARET = "caret";
    public static final String MARGIN = "margin";
    public static final String HIGHLIGHTER = "highlighter";
    public static final String FOCUS_ACCELERATOR_KEY = "focusAcceleratorKey";
    /*properties*******************************/
    private Document model;
    private Caret caret;
    private Highlighter highlighter;
    private boolean opaque;
    private Color caretColor;
    private Color selectionColor;
    private Color selectedTextColor;
    private Color disabledTextColor;
    private boolean editable;
    private Insets margin;
    private char focusAccelerator;
    private Action focusAction = new FocusAction();
    private ButtonManager buttonManager;
    private EditorKit editorKit;
    /*Constructor******************************/
    public TextComponent(EditorKit _ek) {
        editorKit = _ek;
        if (editorKit!=null)
            editorKit.install(this);
        enableEvents(AWTEvent.KEY_EVENT_MASK);
        setEditable(true);
        setLayout(null);
        updateUI();
        ToolTipManager.sharedInstance().registerComponent(this);
    }
    public void setButtonManager(ButtonManager bm) {
        ButtonManager old = buttonManager;
        buttonManager = bm;
        firePropertyChange(BUTTON_MANAGER, old, buttonManager);
    }
    public ButtonManager getButtonManager() {
        return buttonManager;
    }
    /*JComponent******************************/
    public String getToolTipText(MouseEvent event) {
        if (model==null || model.inUpdate())
            return null;
        Point pt = event.getPoint();
        int position = getUI().viewToModel(this, pt);
        if (position==-1)
            return null;
        try {
            return model.getRootElement().getToolTipText(position, Position.Bias.Backward);
        } catch (BadLocationException e) {
            return null;
        }
    }
    /*UI****************************************/
    public TextUI getUI() { return (TextUI)ui; }
    public void setUI(TextUI ui) { super.setUI(ui); }
    protected TextUI createUI() {
        return (TextUI)BasicTextUI.createUI(this);
    }
    public void updateUI() {
        setUI(createUI());
        invalidate();
    }
    public EditorKit getEditorKit() { return editorKit; }
    public void setEditorKit(EditorKit _ek) {
        if (editorKit!=null)
            editorKit.deinstall(this);
        EditorKit oldValue = editorKit;
        editorKit = _ek;
        if (editorKit!=null)
            editorKit.install(this);
        firePropertyChange(EDITOR_KIT, oldValue, editorKit);
    }
    /*Data*****************************************/
    public void setDocument(Document doc) {
        Document old = model;
        model = doc;
        firePropertyChange(DOCUMENT, old, doc);
        revalidate();
        repaint();
    }
    public Document getDocument() {
        return model;
    }
    /*Layout*****************************************/
    public void setMargin(Insets m) {
        Insets old = margin;
        margin = m;
        firePropertyChange(MARGIN, old, m);
        invalidate();
    }
    public Insets getMargin() {
        return margin;
    }
    /*Caret******************************************/
    public Caret getCaret() {
        return caret;
    }
    public void setCaret(Caret c) {
        if (caret!=null) {
            caret.setVisible(false);
            caret.deinstall(this);
        }
        Caret oldCaret = caret;
        caret = c;
        if (caret!=null) {
            caret.install(this);
            caret.setVisible(true);
        }
        firePropertyChange(CARET, oldCaret, caret);
    }
    /*Highlighter****************************************/
    public Highlighter getHighlighter() {
        return highlighter;
    }
    public void setHighlighter(Highlighter h) {
        if (highlighter != null) {
            highlighter.deinstall(this);
        }
        Highlighter old = highlighter;
        highlighter = h;
        if (highlighter != null) {
            highlighter.install(this);
        }
        firePropertyChange(HIGHLIGHTER, old, h);
    }
    /*Colors*****************************************/
    public Color getCaretColor() {
        return caretColor;
    }
    public void setCaretColor(Color c) {
        Color old = caretColor;
        caretColor = c;
        firePropertyChange("caretColor", old, caretColor);
    }
    public Color getSelectionColor() {
        return selectionColor;
    }
    public void setSelectionColor(Color c) {
        Color old = selectionColor;
        selectionColor = c;
        firePropertyChange("selectionColor", old, selectionColor);
    }
    public Color getSelectedTextColor() {
        return selectedTextColor;
    }
    public void setSelectedTextColor(Color c) {
        Color old = selectedTextColor;
        selectedTextColor = c;
        firePropertyChange("selectedTextColor", old, selectedTextColor);
    }
    public Color getDisabledTextColor() {
        return disabledTextColor;
    }
    public void setDisabledTextColor(Color c) {
        Color old = disabledTextColor;
        disabledTextColor = c;
        firePropertyChange("disabledTextColor", old, disabledTextColor);
    }
    /*Data**********************************/
    public void replaceSelection(String content, String format) {
        Document doc = getDocument();
        if (doc!=null) {
            try {
                doc.beginUpdate();
                final int dot = caret.getDot();
                final int mark = caret.getMark();
                final int p0 = Math.min(dot, mark);
                final int p1 = Math.max(dot, mark);
                if (p0!=p1)
                    doc.remove(p0, p1-p0);
                if (content!=null && content.length()>0)
                    doc.insertString(p0, content, format);
            } catch (BadLocationException e) {
                getToolkit().beep();
            } finally {
                doc.endUpdate();
            }
        }
    }
    public void replaceSelection(Element el) {
        Document doc = (Document)getDocument();
        if (doc != null)
            try {
                doc.beginUpdate();
                final int dot = caret.getDot();
                final int mark = caret.getMark();
                final int p0 = Math.min(dot, mark);
                final int p1 = Math.max(dot, mark);
                if (p0!=p1)
                    doc.remove(p0, p1 - p0);
                if (el!=null) {
                    DocumentEvent evt = new DocumentEvent(doc);
                    doc.getRootElement().pasteElementAt(p0, el, evt);
                    doc.fireDocumentChanged(evt);
                }
            } catch (BadLocationException e) {
                getToolkit().beep();
            } finally {
                doc.endUpdate();
            }
    }
    /*Accelerator****************************/
    public void setFocusAccelerator(char aKey) {
        aKey = Character.toUpperCase(aKey);
        KeyStroke[] keyStrokes = getRegisteredKeyStrokes();
        for (int i=0; i<keyStrokes.length; i++) {
            if (getActionForKeyStroke(keyStrokes[i])==focusAction) {
                if (keyStrokes[i].getKeyChar()==aKey)
                    return;
                else
                    unregisterKeyboardAction(keyStrokes[i]);
                break;
            }
        }
        if (aKey!='\0')
            registerKeyboardAction(focusAction, 
                                   KeyStroke.getKeyStroke(aKey, ActionEvent.ALT_MASK),
                                   JComponent.WHEN_IN_FOCUSED_WINDOW);
        char old = focusAccelerator;
        focusAccelerator = aKey;
        firePropertyChange(FOCUS_ACCELERATOR_KEY, old, focusAccelerator);
    }
    public char getFocusAccelerator() {
        return focusAccelerator;
    }
    /*JComponent************************************/
    public boolean isOpaque() { return opaque; }
    public void setOpaque(boolean o) { opaque = o; }
    public void setEnabled(boolean b) {
        super.setEnabled(b);
        repaint();
    }
    public boolean isFocusTraversable() {
        return isEnabled();
    }
    protected void processComponentKeyEvent(KeyEvent evt)  {
        switch (evt.getID()) {
        case KeyEvent.KEY_TYPED:
            Action action = getUI().getEditorKit(this).getDefaultAction();
            if (action!=null) {
                ActionEvent ae = new ActionEvent
                    (this, ActionEvent.ACTION_PERFORMED, 
                     String.valueOf(evt.getKeyChar()), evt.getModifiers());
                action.actionPerformed(ae);
                evt.consume();
            }
            break;
        }
        if (isManagingFocus())
            if (evt.getKeyCode()==KeyEvent.VK_TAB || evt.getKeyChar()=='\t')
                evt.consume();
    }
    /*Editable*************************/
    public boolean isEditable() {
        return editable;
    }
    public void setEditable(boolean b) {
        if (b!=editable) {
            boolean oldVal = editable;
            editable = b;
            firePropertyChange
                ("editable", new Boolean(oldVal), new Boolean(editable));
            repaint();
        }
    }
    /*Scrollable*******************************/
    public Dimension getPreferredScrollableViewportSize() {
        return getPreferredSize();
    }
    public int getScrollableUnitIncrement(Rectangle visibleRect, int orientation, int direction) {
        switch(orientation) {
        case SwingConstants.VERTICAL:
            return visibleRect.height / 10;
        case SwingConstants.HORIZONTAL:
            return visibleRect.width / 10;
        default:
            throw new IllegalArgumentException("Invalid orientation: " + orientation);
        }
    }
    public int getScrollableBlockIncrement(Rectangle visibleRect, int orientation, int direction) {
        switch(orientation) {
        case SwingConstants.VERTICAL:
            return visibleRect.height;
        case SwingConstants.HORIZONTAL:
            return visibleRect.width;
        default:
            throw new IllegalArgumentException("Invalid orientation: " + orientation);
        }
    }  
    public boolean getScrollableTracksViewportWidth() {
        if (getParent() instanceof JViewport) {
            return (((JViewport)getParent()).getWidth() > getPreferredSize().width);
        }
        return false;
    }
    public boolean getScrollableTracksViewportHeight() {
        if (getParent() instanceof JViewport) {
            return (((JViewport)getParent()).getHeight() > getPreferredSize().height);
        }
        return false;
    }
    /*Class*************************************/
    class FocusAction extends AbstractAction {
        public void actionPerformed(ActionEvent e) {
            requestFocus();
        }
        public boolean isEnabled() {
            if (isEditable())
                return true;
            else
                return false;
        }
    }
}
