/*
 *
 *  MakeLib
 *  Copyright (c) 1997-2001 Max Kellermann (max.kellermann@epost.de)
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation;
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: CompositeView.java,v 1.4 2001/08/20 09:23:39 max Exp $
 *
 */

package make.swing.text.view;

import java.awt.Component;
import java.awt.Rectangle;
import java.awt.Shape;
import javax.swing.text.BadLocationException;
import javax.swing.text.Position;
import make.util.Interval;
import make.util.Debug;
import make.swing.text.Element;
import make.swing.text.Document;
import make.swing.text.DocumentEvent;
import make.swing.text.ContainerChange;

public abstract class CompositeView extends View {
    /***properties********************************/
    private static View[] ZERO = new View[0];
    private View[] children;
    private int nchildren;
    private short left;
    private short right;
    private short top;
    private short bottom;
    private Rectangle childAlloc;
    /***Constructor*************************************/
    public CompositeView(Element elem) {
	super(elem);
	children = new View[1];
	nchildren = 0;
	childAlloc = new Rectangle();
    }
    /***CompositeView***********************************/
    protected void loadChildren(ViewFactory f) {
	Element e = getElement();
	int n = e.getElementCount();
	if (n>0) {
	    View[] added = new View[n];
	    for (int i=0; i<n; i++)
		added[i] = f.create(e.getElementAtIndex(i));
	    replace(0, 0, added);
	}
    }
    public void removeAll() {
	replace(0, nchildren, ZERO);
    }
    public void remove(int i) {
	replace(i, 1, ZERO);
    }
    public void insert(int offs, View v) {
	View[] one = new View[] {v};
	replace(offs, 0, one);
    }
    public void append(View v) {
	View[] one = new View[] {v};
	replace(nchildren, 0, one);
    }
    public synchronized void replace(int offset, int length, View[] views) {
	// clear
	for (int i=offset; i<offset+length; i++) {
	    children[i].setParent(null);
	    children[i] = null;
	}
	// update array
	final int delta = views.length - length;
	int src = offset + length;
	int nmove = nchildren - src;
	int dest = src + delta;
	if ((nchildren+delta)>=children.length) {
	    // grow
	    int newLength = Math.max(2*children.length, nchildren + delta);
	    View[] newChildren = new View[newLength];
	    System.arraycopy(children, 0, newChildren, 0, offset);
	    System.arraycopy(views, 0, newChildren, offset, views.length);
	    System.arraycopy(children, src, newChildren, dest, nmove);
	    children = newChildren;
	} else {
	    System.arraycopy(children, src, children, dest, nmove);
	    System.arraycopy(views, 0, children, offset, views.length);
	}
	nchildren = nchildren + delta;
	// parent
	for (int i=0; i<views.length; i++)
	    views[i].setParent(this);
    }
    public int getViewCount() { return nchildren; }
    public View getView(int n) { return children[n]; }
    public Shape getChildAllocation(int index, Shape a) {
	Rectangle alloc = getInsideAllocation(a);
	childAllocation(index, alloc);
	return alloc;
    }
    /***Pool**************************************************/
    protected void replacePool(int first, int count, View[] views) {
	replace(first, count, views);
    }
    public int getPoolViewCount() { return getViewCount(); }
    public View getPoolView(int idx) { return getView(idx); }
    public Shape getPoolViewAllocation(int idx, Shape a) {
	return getChildAllocation(idx, a);
    }
    public int getPoolViewIndexAt(int pos, Position.Bias bias) {
	return getElement().getIndexAt(pos, bias);
    }
    /***update************************************************/
    protected final void updateChildren(ContainerChange ec, ViewFactory f) {
	Element el = getElement();
	Interval removed = ec.getRemoved();
	Interval added = ec.getAdded();
	if (removed!=null && removed.isValid()) {
	    final int first = removed.getFirst();
	    final int count = removed.getCount();
	    if (Debug.DEBUG) {
		if (first<0)
		    throw new InternalError(this+": wrong removed base: "+first);
		if ((first+count)>getPoolViewCount()) {
		    System.err.println("ERROR:");
		    ec.dump();
		    throw new InternalError(this+": wrong removed end: "+first+
					    " count="+removed.getCount()+" vcount="+getPoolViewCount());
		}
	    }
	    replacePool(first, count, ZERO);
	}
	if (added!=null && added.isValid()) {
	    final int first = added.getFirst();
	    final int count = added.getCount();
	    View[] a = new View[count];
	    for (int z=0; z<count; z++)
		a[z] = f.create(el.getElementAtIndex(first+z));
	    replacePool(first, 0, a);
	}
	if (Debug.DEBUG && el.getElementCount()!=getPoolViewCount()) {
	    System.err.println("FATAL ERROR: wrong event");
	    System.err.println("el="+el+" this="+this);
	    System.err.println("el.getElementCount() = "+el.getElementCount());
	    System.err.println("new poolViewCount() = "+getPoolViewCount());
	    System.err.println("added="+added);
	    System.err.println("removed="+removed);
	    System.err.println("changed="+ec.getChanged());
	    throw new InternalError("wrong event");
	}
    }
    protected void forwardUpdate(DocumentEvent evt, ContainerChange chg, 
				 Shape a, ViewFactory f) {
	Element el = getElement();
	if (chg==null) {
	    final int count = getPoolViewCount();
	    for (int z=0; z<count; z++)
		getPoolView(z).documentChanged(evt, getPoolViewAllocation(z, a), f);
	} else {
	    int holeFirst=-1, holeLast=-1;
	    Interval added = chg.getAdded();
	    if (added!=null && added.isValid()) {
		holeFirst = added.getFirst();
		holeLast = added.getLast();
	    }
	    Interval changed = chg.getChanged();
	    if (changed!=null && changed.isValid()) {
		final int first = changed.getFirst();
		final int last = changed.getLast();
		if (Debug.DEBUG && (first<0 || last>=getPoolViewCount())) {
		    System.err.println("FATAL ERROR: wrong changed interval: "+changed);
		    System.err.println("this = "+this);
		    System.err.println("poolViewCount = "+getPoolViewCount());
		    System.err.println("removed="+chg.getRemoved()+" added="+chg.getAdded());
		    System.err.println("changedFirst="+first+" changedLast="+last);
		    throw new InternalError();
		}
		for (int i=first; i<=last; i++)
		    if (i<holeFirst || i>holeLast)
			getPoolView(i).documentChanged(evt, getPoolViewAllocation(i, a), f);
	    }
	}
    }
    protected void updateLayout(DocumentEvent evt, ContainerChange chg, Shape a) {
	preferenceChanged(null, true, true);
	if (a!=null && chg!=null && (chg.getAdded().isValid() ||
				     chg.getRemoved().isValid())) {
	    Rectangle alloc = getInsideAllocation(a);
	    Component host = getContainer();
	    host.repaint(alloc.x, alloc.y, alloc.width, alloc.height);
	}
    }
    public boolean isVertical() {
	return getViewCount()==1 && View.isVertical(getView(0));
    }
    /***View***********************************************/
    public void setParent(View parent) {
	super.setParent(parent);
	if (parent!=null && nchildren==0) {
	    ViewFactory f = getViewFactory();
	    loadChildren(f);
	}
    }
    public Shape modelToView(int pos, Shape a) throws BadLocationException {
	if (getViewCount()==0)
	    return a;
	int vIndex = getViewIndexAtPosition(pos);
	if (vIndex!=-1) {
	    View v = getView(vIndex);
	    if (Debug.DEBUG && (pos<v.getStartOffset() || pos>v.getEndOffset())) {
		System.err.println(this+" wrong getViewIndexAtPosition(): pos="+pos+" v="+v);
		for (int z=0; z<getViewCount(); z++)
		    System.err.println("view "+z+": "+getView(z));
	    }
	    Shape retShape = v.modelToView(pos, getChildAllocation(vIndex, a));
	    if (retShape==null && v.getEndOffset()==pos && ++vIndex<getViewCount()) {
		v = getView(vIndex);
		retShape = v.modelToView(pos, getChildAllocation(vIndex, a));
	    }
	    return retShape;
	}
	if (Debug.DEBUG) {
	    Thread.dumpStack();
	    System.err.println("start="+getStartOffset()+" end="+
			       getEndOffset()+" pos="+pos+" idx="+vIndex+" cnt="+getViewCount()+
			       " cnt2="+getElement().getElementCount()+" this="+this);
	}
	throw new BadLocationException("Position not represented by view", pos);
    }
    public Shape modelToView(int p0, int p1, Shape a) 
	throws BadLocationException {
	if (p0<=getStartOffset() && p1>=getEndOffset())
	    return a;
	Rectangle alloc = getInsideAllocation(a);
	Rectangle r0 = new Rectangle(alloc);
	View v0 = getViewAtPosition(p0, r0);
	Rectangle r1 = new Rectangle(alloc);
	View v1 = getViewAtPosition(p1, r1);
	if (v0==v1) {
	    if (v0==null)
		return a;
	    return v0.modelToView(p0, p1, a);
	}
	final int viewCount = getViewCount();
	for (int counter=0; counter<viewCount; counter++) {
	    View v = getView(counter);
	    if (v==v0 || v==v1) {
		View endView;
		Rectangle retRect;
		Rectangle tempRect = new Rectangle();
		if (v==v0) {
		    retRect = v0.modelToView(p0, v0.getEndOffset(), a).getBounds();
		    endView = v1;
		} else {
		    retRect = v1.modelToView(v1.getStartOffset(), p1, a).getBounds();
		    endView = v0;
		}
		// entirely covered by range
		while (++counter<viewCount && (v=getView(counter))!=endView) {
		    tempRect.setBounds(alloc);
		    childAllocation(counter, tempRect);
		    retRect.add(tempRect);
		}
		if (endView!=null) {
		    Shape endShape = endView==v1
			?v1.modelToView(v1.getStartOffset(), p1, r1)
			:v0.modelToView(p0, v0.getEndOffset(), r0);
		    retRect.add(endShape instanceof Rectangle
				?(Rectangle)endShape
				:endShape.getBounds());
		}
		return retRect;
	    }
	}
	throw new BadLocationException("Position not represented by view", p0);
    }
    public int viewToModel(float x, float y, Shape a) {
	Rectangle alloc = getInsideAllocation(a);
	if (isBefore((int)x, (int)y, alloc)) {
	    int retValue = -1;
	    try {
		retValue = getNextVisualPositionFrom(-1, EAST);
	    } catch (BadLocationException ble) { 
	    } catch (IllegalArgumentException iae) { 
	    }
	    if (retValue==-1)
		retValue = getStartOffset();
	    return retValue;
	} else if (isAfter((int)x, (int)y, alloc)) {
	    int retValue = -1;
	    try {
		retValue = getNextVisualPositionFrom(-1, WEST);
	    } catch (BadLocationException ble) {
	    } catch (IllegalArgumentException iae) { 
	    }
	    if (retValue==-1)
		retValue = getEndOffset() - 1;
	    return retValue;
	} else {
	    View v = getViewAtPoint((int)x, (int)y, alloc);
	    return v==null
		?-1
		:v.viewToModel(x, y, alloc);
	}
    }
    public int getRowStart(int pos) {
	View v = getViewAtPosition(pos, null);
	return v==null?-1:v.getRowStart(pos);
    }
    public int getRowEnd(int pos) {
	View v = getViewAtPosition(pos, null);
	return v==null?-1:v.getRowEnd(pos);
    }
    public int getNextVisualPositionFrom(int pos, int direction) 
	throws BadLocationException {
	switch (direction) {
	case NORTH:
	case SOUTH:
	    return getNextNorthSouthVisualPositionFrom(pos, direction);
	case EAST:
	case WEST:
	    return getNextEastWestVisualPositionFrom(pos, direction);
	default:
	    throw new IllegalArgumentException("Bad direction: " + direction);
	}
    }
    public void documentChanged(DocumentEvent evt, Shape a, ViewFactory f) {
	ContainerChange chg = evt.getContainerChange(getElement());
	if (chg!=null)
	    updateChildren(chg, f);
	forwardUpdate(evt, chg, a, f);
	updateLayout(evt, chg, a);
    }
    /***Abstract******************************************************/
    protected abstract boolean isBefore(int x, int y, Rectangle alloc);
    protected abstract boolean isAfter(int x, int y, Rectangle alloc);
    protected abstract View getViewAtPoint(int x, int y, Rectangle alloc);
    protected abstract void childAllocation(int index, Rectangle a);
    /***Find****************************************************/
    protected int getViewIndexAtPosition(int offset) {
	if (offset<getStartOffset() || offset>getEndOffset())
	    return -1;
	int lower=0, upper=getViewCount()-1, mid=0;
	int p0=getStartOffset();
	if (upper==-1 || offset>=(getEndOffset()-1))
	    return upper;
	if (offset<=p0)
	    return lower;
	while (lower<=upper) {
	    mid = (upper+lower)/2;
	    View elem = getView(mid);
	    p0 = elem.getStartOffset(); 
	    int p1 = elem.getEndOffset();
	    if (offset>=p0 && offset<p1)
		return mid;
	    else if (offset<p0)
		upper = mid - 1;
	    else
		lower = mid + 1;
	}
	return offset<p0
	    ?mid-1
	    :mid;
    }
    protected View getViewAtPosition(int pos, Rectangle a) {
	int index = getViewIndexAtPosition(pos);
	if (index==-1)
	    return null;
	if (a!=null)
	    childAllocation(index, a);
	return getView(index);
    }
    /***Layout****************************************************/
    protected Rectangle getInsideAllocation(Shape a) {
	if (a==null) return null;
	Rectangle alloc = a instanceof Rectangle
	    ?(Rectangle)a:a.getBounds();
	childAlloc.setBounds(alloc);
	childAlloc.x += left;
	childAlloc.y += top;
	childAlloc.width -= left + right;
	childAlloc.height -= top + bottom;
	return childAlloc;
    }
    protected final void setParagraphInsets(Element attr) {
	top = 4;
	bottom = 4;
	left = 0;
	right = 0;
    }
    protected final void setInsets(short top, short left, short bottom, short right) {
	this.top = top;
	this.left = left;
	this.right = right;
	this.bottom = bottom;
    }
    protected final short getLeftInset() { return left; }
    protected final short getRightInset() { return right; }
    protected final short getTopInset() { return top; }
    protected final short getBottomInset() { return bottom; }
    /***CompositeView**********************************************/
    protected int getNextNorthSouthVisualPositionFrom(int pos, int direction)
	throws BadLocationException {
	if (getViewCount()==0)
	    return pos==-1 ? getStartOffset() : -1;
	final boolean isNorth = (direction == NORTH);
	if (pos == -1) {
	    View v = (isNorth) ? getView(getViewCount() - 1) : getView(0);
	    return v.getNextVisualPositionFrom(pos, direction);
	} else {
	    int retValue;
	    int vIndex = getViewIndexAtPosition(pos);
	    View v = getView(vIndex);
	    retValue = v.getNextVisualPositionFrom(pos, direction);
	    if (retValue==-1) {
		if ((isNorth && --vIndex>=0) || (!isNorth && ++vIndex<getViewCount())) {
		    v = getView(vIndex);
		    retValue = v.getNextVisualPositionFrom(-1, direction);
		}
	    }
	    return retValue;
	}
    }
    protected int getNextEastWestVisualPositionFrom(int pos, int direction)
	throws BadLocationException {
	if (getViewCount()==0)
	    return pos==-1 ? getStartOffset() : -1;
	boolean isEast = direction==EAST;
	int retValue;
	int increment = isEast?1:-1;
	if (pos==-1) {
	    View v = isEast ? getView(0) : getView(getViewCount() - 1);
	    retValue = v.getNextVisualPositionFrom(pos, direction);
	    if (retValue==-1 && isEast && getViewCount()>1) {
		v = getView(1);
		retValue = v.getNextVisualPositionFrom(-1, direction);
	    }
	} else {
	    int vIndex = getViewIndexAtPosition(pos);
	    View v = getView(vIndex);
	    retValue = v.getNextVisualPositionFrom(pos, direction);
	    if (retValue==-1) {
		vIndex += increment;
		if (vIndex>=0 && vIndex<getViewCount()) {
		    v = getView(vIndex);
		    retValue = v.getNextVisualPositionFrom(-1, direction);
		    if (retValue==pos)
			retValue = v.getNextVisualPositionFrom(retValue, direction);
		}
	    } else {
		vIndex += increment;
		if (((increment==1 && v.getEndOffset()==retValue) ||
		     (increment==-1 && v.getStartOffset()==retValue)) &&
		    vIndex>=0 && vIndex<getViewCount()) {
		    v = getView(vIndex);
		    retValue = v.getNextVisualPositionFrom(-1, direction);
		}
	    }
	}
	return retValue;
    }
}
