/*
 *
 *  MakeLib
 *  Copyright (c) 1997-2001 Max Kellermann (max.kellermann@epost.de)
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation;
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: LabelView.java,v 1.8 2001/10/18 21:05:20 max Exp $
 *
 */

package make.swing.text.view;

import java.awt.Shape;
import java.awt.Rectangle;
import java.awt.Font;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.font.FontRenderContext;
import java.awt.font.GlyphVector;
import java.awt.font.LineMetrics;
import javax.swing.text.Segment;
import javax.swing.text.BadLocationException;
import make.util.Debug;
import make.swing.text.Element;
import make.swing.text.Document;
import make.swing.text.DocumentEvent;
import make.swing.text.TextComponent;
import make.swing.text.TextUI;

public abstract class LabelView extends View {
    public final class Style {
	/*properties*************************/
	protected Font font;
	protected Color fg;
	protected Color bg;
	protected boolean underline, strike, superscript, subscript;
	protected FontRenderContext fontRenderContext;
	protected Segment text;
	protected GlyphVector glyphVector;
	protected LineMetrics lineMetrics;
	/*Constructor*****************************/
	public void loadPropertiesFromAttributes(Document doc, Element attr) {
	    TextComponent target = getContainer();
	    TextUI ui = target.getUI();
	    // extract misc info
	    font = ui.getFont(target, attr);
	    fg = (Color)attr.getAttribute("textcolor");
	    if (fg==null) fg = Color.black;
	    bg = (Color)attr.getAttribute("bgcolor");
	    underline = attr.containsAttribute("underline", Boolean.TRUE);
	    strike = attr.containsAttribute("strikethrough", Boolean.TRUE);
	    superscript = attr.containsAttribute("superscript", Boolean.TRUE);
	    subscript = attr.containsAttribute("subscript", Boolean.TRUE);
	    // get font metrics
	    Graphics2D g2d = (Graphics2D)target.getGraphics();
	    try {
		fontRenderContext = g2d.getFontRenderContext();
	    } finally {
		if (g2d!=null)
		    g2d.dispose();
	    }
	}
	public void updateText(Segment _text) {
	    text = _text;
	    glyphVector = font.createGlyphVector(fontRenderContext, text);
	    lineMetrics = font.getLineMetrics(text.toString(), fontRenderContext);
	}
	/*Formatinfo******************************************/
	public float getTextWidth(int offset) {
	    if (Debug.DEBUG) {
		if (offset>text.count)
		    throw new InternalError("invalid offset: ofs="+offset+
					    " count="+text.count);
		if ((text.offset+text.count)>text.array.length)
		    throw new InternalError("invalid segment: len="+text.array.length+
					    " ofs="+text.offset+" count="+text.count);
	    }
	    int nextX = 0;
	    return (float)font.getStringBounds(text, text.offset, text.offset+offset, fontRenderContext).getWidth();
	}
	public float getTextWidth() {
	    return getTextWidth(text.count);
	}
	public Color getSelectionColor(TextComponent c) {
	    if (c==null) return null;
	    Color sc = c.getSelectedTextColor();
	    return sc==null || sc.equals(fg)?null:sc;
	}
	public float getFontHeight() { 
	    return (float)font.getStringBounds
		(text.toString(), fontRenderContext).getHeight();
	}
	public float getFontDescent() { 
	    return (float)lineMetrics.getDescent();
	}
	/*Search******************************************/
	public int getTextOffset(int x, boolean round) {
	    float currX=0, nextX=currX;
	    final int n = text.offset + text.count;
	    for (int i=text.offset; i<n; i++) {
		nextX += (float)font.getStringBounds
		    (""+text.array[i], fontRenderContext).getWidth();
		if (x>=currX && x<nextX)
		    if (round==false || ((x-currX)<(nextX-x)))
			return i - text.offset;
		    else
			return i + 1 - text.offset;
		currX = nextX;
	    }
	    return text.count;
	}
	public int getTextOffset(int x) {
	    return getTextOffset(x, true);
	}
	/*Draw***********************************************/
	public int drawText(int x, int y, Graphics g) {
	    if (Debug.DEBUG && text.offset+text.count>text.array.length) {
		getDocument().getRootElement().dump(System.err);
		throw new InternalError(LabelView.this+".drawText len="+text.array.length+
					" ofs="+text.offset+" count="+text.count);
	    }
	    if (text.count>0)
		g.drawChars(text.array, text.offset, text.count, x, y);
	    int nextX = x;
	    nextX += (int)font.getStringBounds
		(text, text.offset, text.offset+text.count, fontRenderContext).getWidth();
	    return nextX;
	}
	public int paintText(int x, int y, Graphics g, 
			     Color color, boolean stripWS) {
	    final int startX = x;
	    g.setColor(color);
	    if (superscript)
		y -= lineMetrics.getAscent() / 2;
	    else if (subscript)
		y += lineMetrics.getDescent();
	    x = drawText(x, y, g);
	    if (underline || strike) {
		int width = x-startX;
		if (stripWS) {
		    while (text.count>0 && 
			   (Character.isWhitespace(text.array[text.count-1]) || 
			    text.array[text.count-1]=='\u00b7')) {
			text.count -= 1;
		    }
		    width = (int)font.getStringBounds
			(text, text.offset, text.offset+text.count, fontRenderContext).getWidth();
		}
		if (underline)
		    y += 1;
		else if (strike)
		    y -= lineMetrics.getAscent()/2;
		g.drawLine(startX, y, x, y);
	    }
	    return x;
	}
	public int paintText(int x, int y, Graphics g, 
			     boolean stripWS) {
	    return paintText(x, y, g, fg, stripWS);
	}
	public void paintBackground(Graphics g, Shape a) {
	    if (bg!=null) {
		Rectangle alloc = (a instanceof Rectangle)
		    ?(Rectangle)a
		    :a.getBounds();
		g.setColor(bg);
		g.fillRect(alloc.x, alloc.y, alloc.width, alloc.height);
	    }
	}
	/*Break******************************************/
	public int getBreakLocation(int x) {
	    final int index = getTextOffset(x, false);
	    final int first = text.offset + Math.min(index, text.count-1);
	    for (int i=first; i>=text.offset; i--) {
		final char ch = text.array[i];
		if (Character.isWhitespace(ch) || ch=='\u00b7')
		    return i-text.offset+1;
	    }
	    return index;
	}
	public int getBreakWeight(int offset, float pos, float len) {
	    // TODO : offset beachten
	    final int index = getTextOffset((int)len);
	    if (index==0)
		return BadBreakWeight;
	    final int first = text.offset+Math.min(index, text.count-1);
	    for (int i=first; i>=text.offset; i--) {
		final char ch = text.array[i];
		if (Character.isWhitespace(ch) || ch=='\u00b7')
		    return ExcellentBreakWeight;
	    }
	    return GoodBreakWeight;
	}
    }
    /*properties*********************************/
    private int allocX, allocY;
    protected Segment text = new Segment();
    protected Style style = new Style();
    protected boolean styleMaster = true;
    /*Constructor************************************/
    public LabelView(Element elem) { super(elem); }
    /*LabelView*********************************/
    protected void loadText() {
	text.offset = 0; text.count = text.array.length;
    }
    protected abstract void loadTextFromElement();
    /*Misc***********************************************/
    public Segment getText() { return text; }
    protected final void sync() {
	if (styleMaster && style.font==null)
	    style.loadPropertiesFromAttributes(getDocument(), getElement());
	if (text.array==null) {
	    loadTextFromElement();
	    loadText();
	    style.updateText(text);
	}
	loadText();
	if (Debug.DEBUG) {
	    if (text.array==null)
		throw new InternalError();
	    if (text.offset<0)
		throw new InternalError();
	    if ((text.offset+text.count)>text.array.length)
		throw new InternalError("this="+this+" offs="+text.offset+" count="+text.count+" arraylen="+text.array.length);
	}
    }
    public Style getStyle() { return style; }
    public int getTextOffset(int relX, boolean round) {
	return getStartOffset()+style.getTextOffset(relX, round);
    }
    /*View*************************************************/
    public View createFragment(int p0, int p1) {
	return new LabelFragment(this, p0, p1);
    }
    public void documentChanged(DocumentEvent e, Shape a, ViewFactory f) {
	text.array = null;
	style.font = null;
	sync();
    }
    public float getMaximumSpan(int axis) {
	return axis==Y_AXIS
	    ?getPreferredSpan(axis)
	    :super.getMaximumSpan(axis);
    }
    public float getPreferredSpan(int axis) {
	sync();
	if (axis==X_AXIS) {
	    final float width = style.getTextWidth();
	    return Math.max(width, 1.0f);
	} else
	    return style.getFontHeight();
    }
    public Shape modelToView(int pos, Shape a) throws BadLocationException {
	Rectangle alloc = a instanceof Rectangle
	    ?(Rectangle)a
	    :a.getBounds();
	sync();
	if (pos>=getEndOffset())
	    return new Rectangle(alloc.x+alloc.width, alloc.y, 
				 0, (int)style.getFontHeight());
	float width = style.getTextWidth(pos-getStartOffset());
	return new Rectangle(alloc.x+(int)width, alloc.y, 0, (int)style.getFontHeight());
    }
    public int viewToModel(float x, float y, Shape a) {
	Rectangle alloc = a.getBounds();
	sync();
	final int offs = style.getTextOffset((int)x-alloc.x);
	final int retValue = getStartOffset() + offs;
	return retValue;
    }
    public int getNextVisualPositionFrom(int pos, int direction) 
	throws BadLocationException {
	if (DUMP_NEXTVISUAL)
	    System.err.println(this+": nextVisual pos="+pos+" dir="+direction+" end="+getEndOffset());
	switch (direction) {
	case NORTH: 
	case SOUTH:
	    if (pos==-1) {
		sync();
		if (text.count==0) 
		    return getStartOffset();
		final int x = getMagicX()-allocX;
		if (x<0)
		    return getStartOffset();
		else
		    return getTextOffset(x, false);
	    } else
		return -1;
	case EAST:
	    if (pos==-1)
		return getStartOffset();
	    pos++;
	    if (pos>=getEndOffset())
		return -1;
	    if (DUMP_NEXTVISUAL)
		System.err.println(this+": nextMajor res="+pos);
	    return pos;
	case WEST:
	    if (pos==-1)
		return getEndOffset()-1;
	    if (pos<=getStartOffset())
		return -1;
	    return pos-1;
	default:
	    throw new IllegalArgumentException("Bad direction: " + direction);
	}
    }
    final int paintText(int x, int y, Graphics g, int p0,
			int p1, boolean stripWS) {
	sync();
	text.count = p1-p0;
	text.offset += p0-getStartOffset();
	return style.paintText(x, y, g, stripWS);
    }
    final int paintText(int x, int y, Graphics g, int p0,
			int p1, Color color, boolean stripWS) {
	sync();
	text.count = p1-p0;
	text.offset += p0-getStartOffset();
	return style.paintText(x, y, g, color, stripWS);
    }
    protected void paintText(Graphics g, Shape a,
			     boolean stripWhitespace) {
	final int p0 = getStartOffset();
	final int p1 = getEndOffset();
	Rectangle alloc = a.getBounds();
	sync();
	int y = alloc.y + alloc.height - (int)style.getFontDescent();
	g.setFont(style.font);
	paintText(alloc.x, y, g, p0, p1, stripWhitespace);
    }
    /*View***************************************************/
    public void paint(Graphics g, Shape a) {
	Rectangle alloc = a.getBounds();
	allocX = alloc.x; allocY = alloc.y;
	style.paintBackground(g, a);
	TextComponent tc = getContainer();
	if (tc!=null) {
	    Highlighter h = tc.getHighlighter();
	    h.paintHighlights(g, getStartOffset(), getEndOffset(), 
			      a, tc, this);
	}
	paintText(g, a, false);
    }
    public int getBreakWeight(int axis, int offset, float pos, float len) {
	if (axis==X_AXIS) {
	    loadTextFromElement(); sync();
	    style.updateText(text);
	    return style.getBreakWeight(offset, pos, len);
	} else
	    return super.getBreakWeight(axis, offset, pos, len);
    }
    public View breakView(int axis, int p0, float pos, float len) {
	if (axis==X_AXIS) {
	    sync();
	    final int d = p0 - getStartOffset();
	    text.count -= d;
	    text.offset += d;
	    int p1 = p0+style.getBreakLocation((int)len);
	    return createFragment(p0, p1);
	}
	return this;
    }
    public float getAlignment(int axis) {
	if (axis == Y_AXIS) {
	    float h = style.getFontHeight();
	    float d = style.getFontDescent();
	    float align = (h - d) / h;
	    return align;
	} 
	return super.getAlignment(axis);
    }
}
