/*
 *
 *  MakeLib
 *  Copyright (c) 1997-2001 Max Kellermann (max.kellermann@epost.de)
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation;
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: DOMUtil.java,v 1.29 2002/01/03 12:23:51 max Exp $
 *
 */

package make.xml;

import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Vector;
import java.util.StringTokenizer;
import java.util.Locale;
import java.io.IOException;
import java.io.InputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.Writer;
import java.io.ByteArrayOutputStream;
import java.io.ByteArrayInputStream;
import java.io.Reader;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.w3c.dom.Node;
import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Text;
import org.w3c.dom.Attr;
import org.w3c.dom.traversal.NodeIterator;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.ContentHandler;
import org.xml.sax.EntityResolver;
import org.xml.sax.XMLReader;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.ParserConfigurationException;
import make.util.Debug;
import make.util.Strings;
import make.util.LocalizedString;

/**
 * This class contains some static helper functions to handle DOM trees.
 */
public abstract class DOMUtil {
    /*CONST********************************/
    public static final Locale locale = Locale.getDefault();
    public static final String language = locale.getLanguage();
    /*static********************************/
    public static final EntityResolver emptyEntityResolver = new EntityResolver() {
            private byte[] emptyArray = {};
            public InputSource resolveEntity(String publicId,
                                             String systemId) {
                return new InputSource(new ByteArrayInputStream(emptyArray));
            }
        };
    protected static DocumentBuilderFactory documentBuilderFactory
        = DocumentBuilderFactory.newInstance();
    protected static SAXParserFactory saxParserFactory
        = SAXParserFactory.newInstance();
    static {
        documentBuilderFactory.setNamespaceAware(true);
        saxParserFactory.setNamespaceAware(true);
    }
    /*Static********************************/
    protected static Node[] toNodeArray(Vector vc) {
        Node[] ret = new Node[vc.size()];
        for (int z=0; z<ret.length; z++)
            ret[z] = (Node)vc.elementAt(z);
        return ret;
    }
    protected static Element[] toElementArray(Vector vc) {
        Element[] ret = new Element[vc.size()];
        for (int z=0; z<ret.length; z++)
            ret[z] = (Element)vc.elementAt(z);
        return ret;
    }
    /**
     * Returns all elements contained in the specified Element.
     */
    public static Element[] getChildElements(Element el) {
        if (el==null) return null;
        NodeList nl = el.getChildNodes();
        final int c = nl.getLength();
        Vector vc = new Vector(c);
        for (int z=0; z<c; z++) {
            Node n = nl.item(z);
            if (n instanceof Element)
                vc.addElement(n);
        }
        return toElementArray(vc);
    }
    /**
     * Returns all elements contained in the specified Element
     * which are in the specified namespace.
     */
    public static Element[] getChildElementsNS(Element el, String namespace) {
        if (el==null) return null;
        NodeList nl = el.getChildNodes();
        final int c = nl.getLength();
        Vector vc = new Vector(c);
        for (int z=0; z<c; z++) {
            Node n = nl.item(z);
            if (n instanceof Element && 
                namespace.equals(n.getNamespaceURI()))
                vc.addElement(n);
        }
        return toElementArray(vc);
    }
    /**
     * Returns all elements contained in the specified Element
     * which have the specified tag.
     */
    public static Element[] getChildElements(Element el, String tag) {
        if (el==null) return null;
        NodeList nl = el.getChildNodes();
        final int c = nl.getLength();
        Vector vc = new Vector(c);
        for (int z=0; z<c; z++) {
            Node n = nl.item(z);
            if (n instanceof Element && 
                tag.equals(n.getNodeName()))
                vc.addElement(n);
        }
        return toElementArray(vc);
    } 
    /**
     * Returns all elements contained in the specified Element
     * which have the specified tag and are in the specified namespace.
     */
    public static Element[] getChildElementsNS(Element el, String namespace,
                                               String tag) {
        if (el==null) return null;
        NodeList nl = el.getChildNodes();
        final int c = nl.getLength();
        Vector vc = new Vector(c);
        for (int z=0; z<c; z++) {
            Node n = nl.item(z);
            if (n instanceof Element && 
                namespace.equals(n.getNamespaceURI()) &&
                tag.equals(n.getLocalName()))
                vc.addElement(n);
        }
        return toElementArray(vc);
    }
    /**
     * Finds all child elements with a specified tag and compares
     * an attribute with the specified value. Only matching
     * elements are returned.
     */
    public static Element getChildElementByAttribute(Element el, String name, 
                                                     String attributeName, 
                                                     String compareTo){
        Element[] children = DOMUtil.getChildElements(el,name);
        for(int i=0;i<children.length;i++){
            if (children[i].getAttribute(attributeName).equals(compareTo))
                return children[i];
        }
        return null;
    }
    /**
     * Tests if an element with the tag exists.
     */
    public static boolean childTagExists(Element el, String tag) {
        return el!=null && getChildElements(el, tag).length>0;
    }
    /**
     * Returns the first element with the specified tag.
     */
    public static Element getUniqueTag(Element el, String tag) {
        if (el==null) return null;
        NodeList nl = el.getChildNodes();
        final int c = nl.getLength();
        for (int z=0; z<c; z++) {
            Node n = nl.item(z);
            if (n instanceof Element && 
                tag.equals(n.getNodeName()))
                return (Element)n;
        }
        return null;
    }
    /**
     * Returns the first element with the specified tag
     * in the given namespace.
     */
    public static Element getUniqueTagNS(Element el, String namespace,
                                         String tag) {
        if (el==null) return null;
        NodeList nl = el.getChildNodes();
        final int c = nl.getLength();
        for (int z=0; z<c; z++) {
            Node n = nl.item(z);
            if (n instanceof Element && 
                namespace.equals(n.getNamespaceURI()) &&
                tag.equals(n.getLocalName()))
                return (Element)n;
        }
        return null;
    }
    public static Element createUniqueTag(Element el, String tag) {
        if (el==null) return null;
        Element ut = getUniqueTag(el, tag);
        if (ut!=null)
            return ut;
        ut = el.getOwnerDocument().createElement(tag);
        el.appendChild(ut);
        return ut;
    }
    public static Element createDeepUniqueTag(Element el, String tag) {
        if (el==null) return null;
        if (tag==null) return el;
        StringTokenizer st = new StringTokenizer(tag, "/");
        while (el!=null && st.hasMoreTokens()) {
            String t = Strings.trim(st.nextToken());
            if (t!=null && !t.equals("."))
                el = createUniqueTag(el, t);
        }
        return el;
    }

    /*Extract strings********************************/
    public static String getElementString(Element el, boolean preserveWS) {
        if (el==null)
            return null;
        Node child = el.getFirstChild();
        if (child==null || !(child instanceof Text))
            return null;
        String data = ((Text)child).getData();
        for (child=child.getNextSibling(); child instanceof Text; 
             child=child.getNextSibling())
            data += ((Text)child).getData();
        if (!preserveWS)
            data = Strings.normalize(data);
        return data;
    }
    public static String getElementString(Element el, String tag,
                                          boolean preserveWS) {
        return getElementString(getUniqueTag(el, tag), preserveWS);
    }
    public static String getElementStringNS(Element el, String namespace,
                                            String tag, boolean preserveWS) {
        return getElementString(getUniqueTagNS(el, namespace, tag), preserveWS);
    }
    public static String getElementParameter(Element el, String name,
                                             boolean preserveWS) {
        String value = getElementString(el, name, preserveWS);
        if (value!=null)
            return value;
        return Strings.trim(el.getAttribute(name));
    }
    public static String getElementParameterNS(Element el, String namespace,
                                               String name, boolean preserveWS) {
        String value = getElementStringNS(el, namespace, name, preserveWS);
        if (value!=null)
            return value;
        return Strings.trim(el.getAttributeNS(namespace, name));
    }
    // Compatibility
    public static String getElementString(Element el) {
        return getElementString(el, false);
    }
    public static String getElementString(Element el, String tag) {
        return getElementString(el, tag, false);
    }
    public static String getElementStringNS(Element el, String namespace,
                                            String tag) {
        return getElementStringNS(el, namespace, tag, false);
    }

    /*Localization******************************/
    public static Element getLocalizedUniqueTag(Element el, String tag,
                                                Locale locale) {
        Element[] ch = getChildElements(el, tag);
        if (ch==null || ch.length==0)
            return null;
        String language = locale.getLanguage();
        for (int z=0; z<ch.length; z++)
            if (language.equals(ch[z].getAttribute("xml:lang")))
                return ch[z];
        return ch[0];
    }
    public static Element getLocalizedUniqueTagNS(Element el, String namespace,
                                                  String tag, Locale locale) {
        Element[] ch = getChildElementsNS(el, namespace, tag);
        if (ch==null || ch.length==0)
            return null;
        String language = locale.getLanguage();
        for (int z=0; z<ch.length; z++)
            if (language.equals(ch[z].getAttribute("xml:lang")))
                return ch[z];
        return ch[0];
    }
    public static String getLocalizedElementString(Element el, String tag,
                                                   Locale locale, boolean preserveWS) {
        return getElementString(getLocalizedUniqueTag(el, tag, locale), preserveWS);
    }
    public static String getLocalizedElementStringNS(Element el, String namespace,
                                                     String tag, Locale locale,
                                                     boolean preserveWS) {
        return getElementString(getLocalizedUniqueTagNS(el, namespace,
                                                        tag, locale),
                                preserveWS);
    }
    public static String getLocalizedElementParameterNS(Element el, String namespace,
                                                        String name, Locale locale,
                                                        boolean preserveWS) {
        String value = getLocalizedElementStringNS
            (el, namespace, name, locale, preserveWS);
        if (value!=null)
            return value;
        return Strings.trim(el.getAttributeNS(namespace, name));
    }
    // LocalizedString
    public static LocalizedString getLocalizedElementString(Element el, String tag,
                                                            boolean preserveWS) {
        Element[] ch = getChildElements(el, tag);
        Map map = new HashMap(ch.length);
        for (int z=0; z<ch.length; z++) {
            String text = getElementString(ch[z], preserveWS);
            if (text==null)
                continue;
            // convert the xml:lang attribute to a Locale object
            Locale locale;
            String localeString = Strings.trim(ch[z].getAttribute("xml:lang"));
            if (localeString!=null) {
                String language = "", country = "", variant="";
                int idx0 = localeString.indexOf('_');
                if (idx0>=0) {
                    language = localeString.substring(0, idx0);
                    int idx1 = localeString.indexOf('_', idx0+1);
                    if (idx1>=0) {
                        country = localeString.substring(idx0+1, idx1);
                        variant = localeString.substring(idx1+1);
                    } else {
                        country = localeString.substring(idx0+1);
                    }
                } else {
                    language = localeString;
                }
                locale = new Locale(language, country, variant);
            } else {
                locale = null;
            }
            map.put(locale, text);
        }
        return LocalizedString.load(map);
    }
    public static void setLocalizedElementString(Element el, String tag,
                                                 LocalizedString ls) {
        // delete old values
        Element[] och = getChildElements(el, tag);
        for (int z=0; z<och.length; z++)
            el.removeChild(och[z]);
        // add new values
        if (ls==null)
            return;
        Document doc = el.getOwnerDocument();
        String defaultText = ls.get(null);
        if (defaultText!=null) {
            Element ch = doc.createElement(tag);
            ch.appendChild(doc.createTextNode(defaultText));
            el.appendChild(ch);
        }
        Iterator it = ls.locales().iterator();
        while (it.hasNext()) {
            Locale locale = (Locale)it.next();
            if (locale==null)
                continue;
            String text = ls.get(locale);
            Element ch = doc.createElement(tag);
            ch.setAttribute("xml:lang", locale.toString());
            ch.appendChild(doc.createTextNode(text));
            el.appendChild(ch);
        }
    }

    /*Misc******************************************/
    public static String getNodeText(Node n) {
        if (n==null)
            return " ";
        if (n instanceof Text)
            return n.getNodeValue();
        if (n.hasChildNodes()) {
            StringBuffer sb = new StringBuffer();
            NodeList nl = n.getChildNodes();
            for (int z=0; z<nl.getLength(); z++)
                sb.append(getNodeText(nl.item(z)));
            return sb.toString();
        }
        return " ";
    }
    public static String getDeepString(Element el, String tag) {
        if (el==null || tag==null) return null;
        StringTokenizer st = new StringTokenizer(tag, "/");
        while (el!=null && st.hasMoreTokens()) {
            String t = Strings.trim(st.nextToken());
            if (t==null || t.equals(".")) continue;
            if (t.startsWith("@")) {
                String attrName = t.substring(1);
                return Strings.trim(el.getAttribute(attrName));
            } else
                el = getUniqueTag(el, t);
        }
        return getElementString(el);
    }
    public static void setDeepString(Node n, String path, String value) {
        value = Strings.trim(value);
        StringTokenizer st = new StringTokenizer(path, "/");
        while (st.hasMoreTokens()) {
            String t = Strings.trim(st.nextToken());
            if (t==null || t.equals(".")) continue;
            if (t.startsWith("@")) {
                String attrName = t.substring(1);
                if (value==null)
                    ((Element)n).removeAttribute(attrName);
                else
                    ((Element)n).setAttribute(attrName, value);
                return;
            } else {
                Element nn = getUniqueTag((Element)n, t);
                if (nn==null) {
                    nn = n.getOwnerDocument().createElement(t);
                    n.appendChild(nn);
                }
                n = nn;
            }
        }
        setElementString((Element)n, value);
    }
    public static void setElementString(Element el, String value) {
        while (el.hasChildNodes())
            el.removeChild(el.getLastChild());
        if (value!=null)
            el.appendChild(el.getOwnerDocument().createTextNode(value));
    }
    public static void removeElements(Element p, String tag) {
        Element[] ch = getChildElements(p, tag);
        for (int z=0; z<ch.length; z++)
            p.removeChild(ch[z]);
    }
    public static void setElementString(Element el, String tag, String value) {
        if (value==null)
            removeElements(el, tag);
        else
            setElementString(createUniqueTag(el, tag), value);
    }
    public static boolean getElementBoolean(Element el) {
        return "yes".equals(getElementString(el));
    }
    public static boolean getElementBoolean(Element el, String tag) {
        return getElementBoolean(getUniqueTag(el, tag));
    }
    /*NodeIterator**********************/
    public static Node[] iteratorToArray(NodeIterator ni) {
        Vector vc = new Vector();
        Node node;
        while ((node=ni.nextNode())!=null)
            vc.addElement(node);
        return (Node[])vc.toArray(new Node[]{});
    }
    public static void appendIterator(Vector dest, NodeIterator src) {
        Node node;
        while ((node=src.nextNode())!=null)
            dest.addElement(node);
    }
    /*Factory***************************************************/
    public static Document createDocument() {
        try {
            DocumentBuilder builder = documentBuilderFactory.newDocumentBuilder();
            return builder.newDocument();
        } catch (ParserConfigurationException e) {
            Debug.report(e);
            throw new InternalError(e.toString());
        }
    }
    public static Element createDocument(String tag) {
        Document doc = createDocument();
        Element root = doc.createElement(tag);
        doc.appendChild(root);
        return root;
    }
    public static Node cloneNode(Document doc, Node n, boolean recursive) {
        Node result;
        switch (n.getNodeType()) {
        case Node.ATTRIBUTE_NODE:
            result = doc.createAttribute(n.getNodeName());
            break;
        case Node.COMMENT_NODE:
            result = doc.createComment(n.getNodeValue());
            break;
        case Node.DOCUMENT_NODE:
            if (Debug.DEBUG && doc!=null)
                throw new InternalError("doc must be null for Document");
            doc = createDocument();
            result = doc;
            break;
        case Node.ELEMENT_NODE: {
            Element el = (Element)n;
            result = doc.createElement(el.getTagName());
            NamedNodeMap as = el.getAttributes();
            final int ac = as.getLength();
            for (int z=0; z<ac; z++)
                ((Element)result).setAttributeNode
                    ((Attr)cloneNode(doc, as.item(z), recursive));
            break;
        }
        case Node.ENTITY_REFERENCE_NODE:
            result = doc.createEntityReference(n.getNodeName());
            break;
        case Node.TEXT_NODE:
            result = doc.createTextNode(n.getNodeValue());
            break;
        default:
            throw new IllegalArgumentException("unknown node: "+n);
        }
        if (recursive) {
            NodeList ch = n.getChildNodes();
            final int cc = ch.getLength();
            for (int z=0; z<cc; z++)
                result.appendChild(cloneNode(doc, ch.item(z), true));
        }
        return result;
    }
    public static Element createStringElement(Document doc, String tag, 
                                              String text) {
        Element el = doc.createElement(tag);
        if (text!=null)
            el.appendChild(doc.createTextNode(text));
        return el;
    }
    public static Element createBooleanElement(Document doc, String tag, 
                                               boolean v) {
        return createStringElement(doc, tag, v?"yes":"no");
    }
    /*Parse****************************************/
    // simple parse
    public static Document parse(InputSource is) 
        throws IOException, SAXException {
        try {
            DocumentBuilder builder = documentBuilderFactory.newDocumentBuilder();
            return builder.parse(is);
        } catch (ParserConfigurationException e) {
            Debug.report(e);
            throw new SAXException(e);
        }
    }
    public static Document parse(Reader is) 
        throws IOException, SAXException {
        return parse(new InputSource(is));
    }
    public static Document parse(InputStream is, String encoding) 
        throws IOException, SAXException {
        return parse(new InputStreamReader(is, encoding));
    }
    public static Document parse(InputStream is) 
        throws IOException, SAXException {
        return parse(new InputSource(is));
    }
    public static Document parse(byte[] file, String encoding) 
        throws IOException, SAXException {
        return parse(new ByteArrayInputStream(file), encoding);
    }
    public static Document parse(File file, String encoding) 
        throws IOException, SAXException {
        return parse(new FileInputStream(file), encoding);
    }
    // parse with dochandler
    public static void parse(InputSource is, ContentHandler dh) 
        throws IOException, SAXException {
        try {
            SAXParser saxParser = saxParserFactory.newSAXParser();
            XMLReader parser = saxParser.getXMLReader();
            parser.setContentHandler(dh);
            parser.parse(is);
        } catch (ParserConfigurationException e) {
            Debug.report(e);
            throw new SAXException(e);
        }
    }
    public static void parse(Reader is, ContentHandler dh) 
        throws IOException, SAXException {
        parse(new InputSource(is), dh);
    }
    public static void parse(InputStream is, String encoding, ContentHandler dh) 
        throws IOException, SAXException {
        parse(new InputStreamReader(is, encoding), dh);
    }
    public static void parse(InputStream is, ContentHandler dh) 
        throws IOException, SAXException {
        parse(new InputSource(is), dh);
    }
    // parse with entity resolver
    public static Document parse(InputSource is, EntityResolver er) 
        throws IOException, SAXException {
        try {
            DocumentBuilder builder = documentBuilderFactory.newDocumentBuilder();
            if (er!=null)
                try {
                    builder.setEntityResolver(er);
                } catch (Throwable e) {
                }
            return builder.parse(is);
        } catch (ParserConfigurationException e) {
            Debug.report(e);
            throw new SAXException(e);
        }
    }
    public static Document parse(Reader is, EntityResolver er) 
        throws IOException, SAXException {
        return parse(new InputSource(is), er);
    }
    public static Document parse(InputStream is, String encoding, EntityResolver er) 
        throws IOException, SAXException {
        return parse(new InputStreamReader(is, encoding), er);
    }
    public static Document parse(InputStream is, EntityResolver er) 
        throws IOException, SAXException {
        return parse(new InputSource(is), er);
    }
    public static Document parse(byte[] file, String encoding, EntityResolver er) 
        throws IOException, SAXException {
        return parse(new ByteArrayInputStream(file), encoding, er);
    }
    /*Serialize****************************************/
    public static byte[] toByteArray(Document doc, String encoding) {
        try {
            ByteArrayOutputStream baos = new ByteArrayOutputStream();
            new DOMWriter(baos, encoding).print(doc);
            return baos.toByteArray();
        } catch (IOException e) {
            Debug.report(e); return null;
        }
    }
    public static byte[] toByteArray(Document doc, String encoding, boolean cfuml) {
        try {
            ByteArrayOutputStream baos = new ByteArrayOutputStream();
            DOMWriter dw = new DOMWriter(baos, encoding);
            dw.setWDDXWorkaround(cfuml);
            dw.print(doc);
            return baos.toByteArray();
        } catch (IOException e) {
            Debug.report(e); return null;
        }
    }
    public static byte[] toByteArray(Document doc, String encoding, DOMWriterFormat dt) {
        try {
            ByteArrayOutputStream baos = new ByteArrayOutputStream();
            DOMWriter dw = new DOMWriter(baos, encoding);
            dw.setFormat(dt);
            dw.print(doc);
            return baos.toByteArray();
        } catch (IOException e) {
            Debug.report(e); return null;
        }
    }
    public static byte[] toByteArray(Document doc, DOMWriterFormat dt, String encoding) {
        try {
            ByteArrayOutputStream baos = new ByteArrayOutputStream();
            DOMWriter dw = new DOMWriter(baos, encoding);
            dw.setFormat(dt);
            dw.print(doc);
            return baos.toByteArray();
        } catch (IOException e) {
            Debug.report(e); return null;
        }
    }
}
