/*
 *
 *  MakeLib
 *  Copyright (c) 1997-2001 Max Kellermann (max.kellermann@epost.de)
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation;
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: DocumentType.java,v 1.10 2001/08/20 09:23:39 max Exp $
 *
 */

package make.xml;

import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;
import java.util.Map;
import java.util.List;
import java.util.HashMap;
import java.util.ArrayList;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.io.IOException;
import org.w3c.dom.Node;
import org.w3c.dom.Attr;
import org.w3c.dom.NodeList;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Element;
import org.w3c.dom.Document;
import org.xml.sax.SAXException;
import javax.xml.transform.TransformerException;
import org.apache.xpath.XPath;
import org.apache.xpath.objects.XObject;
import org.apache.xpath.objects.XNull;
import make.util.Debug;
import make.util.Strings;
import make.io.Streams;

public class DocumentType 
    implements DOMWriterFormat {
    /***static*********************************/
    private static Hashtable pathCache = new Hashtable();
    /***Static********************************/
    public static XPath parseXPath(String path) 
	throws TransformerException {
	XPath pt = (XPath)pathCache.get(path);
	if (pt==null) {
	    pt = XPathUtil.parse(path);
	    pathCache.put(path, pt);
	}
	return pt;
    }
    /***propertes**********************************/
    protected ConfigPath configPath;
    protected String name, publicId, systemId;
    protected Hashtable elements = new Hashtable();
    protected Hashtable lists = new Hashtable();
    protected XPath idPath;
    private Element root;
    /***Constructor*******************************/
    public DocumentType(ConfigPath _cp) 
	throws FileNotFoundException {
	configPath = _cp;
	root = configPath.parse("format.xml");
	if (root==null)
	    throw new FileNotFoundException("format.xml");
	Element head = DOMUtil.getUniqueTag(root, "head");
	name = DOMUtil.getElementString(head, "name");
	publicId = DOMUtil.getElementString(head, "public");
	systemId = DOMUtil.getElementString(head, "system");
	loadStructure(DOMUtil.getUniqueTag(root, "structure"));
	// idPath
	String ip = DOMUtil.getElementString(root, "id-path");
	if (ip!=null)
	    try {
		idPath = parseXPath(ip);
	    } catch (Exception e) {
		e.printStackTrace();
	    }
    }
    protected ElementType loadElementType(Element el) {
	return new ElementType(this, el);
    }
    protected void loadStructure(Element struc) {
	Element[] els = DOMUtil.getChildElements(struc, "element");
	for (int z=0; z<els.length; z++) {
	    ElementType et = loadElementType(els[z]);
	    elements.put(els[z].getAttribute("tag"), et);
	}
    }
    /***Object*******************************/
    public String toString() {
	return "[DocumentType "+name+"]";
    }
    /***Lists*********************************/
    protected Map toMap(Element vd) {
	Element[] ks = DOMUtil.getChildElements(vd, "key");
	if (ks.length==0)
	    ks = DOMUtil.getChildElements(vd, "map-item");
	Map map = new HashMap(ks.length);
	for (int z=0; z<ks.length; z++) {
	    Element k = ks[z];
	    String src = DOMUtil.getElementString(k, "source");
	    String dest;
	    if (src==null) {
		src = Strings.trim(k.getAttribute("source"));
		dest = DOMUtil.getElementString(k);
	    } else {
		dest = DOMUtil.getElementString(k, "dest");
	    }
	    map.put(src, dest);
	}
	return map;
    }
    protected List toList(Element vd) {
	Element[] items = DOMUtil.getChildElements(vd);
	List list = new ArrayList(items.length);
	for (int a=0; a<items.length; a++)
	    list.add(DOMUtil.getElementString(items[a]));
	return list;
    }
    protected Object[] loadSelectMenu(Element vd) {
	Element[] ch = DOMUtil.getChildElements(vd);
	String caption = Strings.trim(vd.getAttribute("caption"));
	final int cc = caption==null?0:1;
	Object[] ret = new Object[ch.length+cc];
	if (caption!=null)
	    ret[0] = caption;
	for (int z=0; z<ch.length; z++)
	    ret[z+cc] = loadSelectMenuItem(ch[z]);
	return ret;
    }
    protected Object loadSelectMenuItem(Element vd) {
	String type = vd.getTagName();
	if (type.equals("menu"))
	    return loadSelectMenu(vd);
	else if (type.equals("item"))
	    return DOMUtil.getElementString(vd);
	else if (type.equals("separator"))
	    return null;
	else {
	    System.err.println("WARNING: unknown menu item: "+type);
	    return null;
	}
    }
    protected Object loadList(Element vd) {
	String tag = vd.getTagName();
	if (tag.equals("lookup-table") || tag.equals("map")) {
	    return toMap(vd);
	} else if (tag.equals("list")) {
	    return toList(vd);
	} else if (tag.equals("menu")) {
	    return (Object[])loadSelectMenuItem(vd);
	} else if (tag.equals("file")) {
	    try {
		String fn = vd.getAttribute("path");
		InputStream is = getClass().getResourceAsStream(fn);
		Document doc = DOMUtil.parse(is, Streams.defaultEncoding);
		Element el = doc.getDocumentElement();
		return loadList(el);
	    } catch (SAXException e) {
		e.printStackTrace();
		return null;
	    } catch (IOException e) {
		e.printStackTrace();
		return null;
	    }
	} else
	    return null;
    }
    protected void loadLists(Element ts) {
	Element[] tsa = DOMUtil.getChildElements(ts);
	for (int z=0; z<tsa.length; z++) {
	    String name = tsa[z].getAttribute("name");
	    Object l = loadList(tsa[z]);
	    if (l==null)
		System.err.println("WARNING: could not load list "+
				   tsa[z].getTagName()+" named "+name);
	    else
		lists.put(name, l);
	}
    }
    /***Embedded********************************/
    protected void getEmbeddedNames(Vector vc, Node n) {
	if (n instanceof Element) {
	    Element el = (Element)n;
	    NamedNodeMap nnm = el.getAttributes();
	    final int ac = nnm.getLength();
	    for (int z=0; z<ac; z++) {
		Attr a = (Attr)nnm.item(z);
		ElementType et = getElementType(el.getTagName());
		if (et==null) continue;
		AttributeType at = et.getAttribute(a.getName());
		if (at==null) continue;
		if (at.getType()==at.T_EMBEDDED) {
		    String name = Strings.trim(a.getValue());
		    if (name!=null && !vc.contains(name))
			vc.addElement(name);
		}
	    }
	}
	NodeList nl = n.getChildNodes();
	final int cc = nl.getLength();
	for (int z=0; z<cc; z++)
	    getEmbeddedNames(vc, nl.item(z));
    }
    public String[] getEmbeddedNames(Node el) {
	Vector vc = new Vector();
	getEmbeddedNames(vc, el);
	String[] n = new String[vc.size()];
	for (int z=0; z<n.length; z++)
	    n[z] = (String)vc.elementAt(z);
	return n;
    }
    /***Info************************************/
    public ConfigPath getConfigPath() { return configPath; }
    public String getName() { return name; }
    public String getPublicId() { return publicId; }
    public String getSystemId() { return systemId; }
    public Enumeration getElementTags() { return elements.keys(); }
    public ElementType getElementType(String tag) {
	return (ElementType)elements.get(tag);
    }
    public Element getMisc(String tag) {
	return DOMUtil.getUniqueTag(root, tag);
    }
    public Object getList(String name) {
	if (Debug.DEBUG && !lists.containsKey(name))
	    System.err.println("DEBUG: list "+name+" does not exist in "+this);
	return lists.get(name);
    }
    public boolean equals(Object o) {
	if (o instanceof DocumentType)
	    return o==this ||
		((DocumentType)o).getName().equals(getName());
	else
	    return super.equals(o);
    }
    public Object extractId(Node el) {
	if (idPath==null)
	    return null;
	try {
	    XObject result = XPathUtil.execute(idPath, el);
	    return result instanceof XNull
		?null
		:Strings.trim(result.str());
	} catch (TransformerException e) {
	    Debug.report(e);
	    return null;
	}
    }
    /***DOMWriterFormat**************************/
    public boolean getPreserveOuterSpaces(Element el) {
	ElementType et = getElementType(el.getTagName());
	if (et==null)
	    return true;
	return !et.isFlowing();
    }
    public boolean isPreformatted(Element el) {
	return false;
    }
    public boolean isFlowing(Element el) {
	ElementType et = getElementType(el.getTagName());
	if (et==null)
	    return true;
	return et.isFlowing();
    }
    public boolean getPreserveInnerSpace(Element el) {
	return true;
    }
    public boolean getPreserveOuterSpace(Element el) {
	return false;
    }
}
