/*
 *
 *  MakeLib
 *  Copyright (c) 1997-2001 Max Kellermann (max.kellermann@epost.de)
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation;
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: RtfToXml.java,v 1.5 2001/08/20 09:23:39 max Exp $
 *
 */

package make.xml;

import org.w3c.dom.*;
import java.io.*;
import java.util.*;
import org.xml.sax.SAXException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.stream.StreamSource;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.dom.DOMResult;
import make.xml.DOMUtil;
import make.xml.DOMWriter;
import make.io.Streams;

interface Format{
}

class ContainerFormat implements Format{
    public String tag;
    public ContainerFormat(Element el){
	tag=el.getAttribute("tag");
    }
}

class ParagraphFormat extends ContainerFormat{
    public ParagraphFormat(Element el){
	super(el);
    }
}

class WhitespaceFormat extends ContainerFormat{
    public WhitespaceFormat(Element el){
	super(el);
    }
}

class IgnoreFormat extends ContainerFormat{
    public IgnoreFormat(Element el){
	super(el);
    }
}

class HeadContainerFormat extends ContainerFormat{
    public HeadContainerFormat(Element el){
	super(el);
    }
}
class EvaluateContainerFormat extends ContainerFormat{
    public EvaluateContainerFormat(Element el){
	super(el);
    }
}

/**
 * This class does basic conversion from RTF (Rich Text Format) to XML.
 * @author Markus Gayda
 */
public class RtfToXml {
    Hashtable  formats = new Hashtable();
    String rootTag;
    String headTag;
    String bodyTag;

    public RtfToXml(Element el){
	loadFormats(DOMUtil.getUniqueTag(el,"formats"));
    }

    public void loadFormats(Element el){
	bodyTag=el.getAttribute("body");
	headTag=el.getAttribute("head");
	rootTag=el.getAttribute("root");
	Element[] ea = DOMUtil.getChildElements(el);
	for(int z=0;z<ea.length;z++){
	    String name = ea[z].getAttribute("name");
	    String type = ea[z].getTagName();
	    if ("container".equals(type))
		formats.put(name,new ContainerFormat(ea[z]));
	    else if ("whitespace".equals(type))
		formats.put(name,new WhitespaceFormat(ea[z]));
	    else if ("ignore".equals(type))
		formats.put(name,new IgnoreFormat(ea[z]));
	    else if ("paragraph".equals(type))
		formats.put(name,new ParagraphFormat(ea[z]));
	    else if ("headcontainer".equals(type))
		formats.put(name,new HeadContainerFormat(ea[z]));
	    else if ("evaluatecontainer".equals(type))
		formats.put(name,new EvaluateContainerFormat(ea[z]));
	    else
		System.err.println("Unknown format: "+type);
	}
    }

    public Document transform(Document doc, File xsl) 
	throws TransformerException {
	TransformerFactory tFactory = TransformerFactory.newInstance();
	Transformer transformer = tFactory.newTransformer(new StreamSource(xsl));
	Document target = DOMUtil.createDocument();
	transformer.transform(new DOMSource(doc),
			      new DOMResult(target));
	return target;
    }

    public char hexStringToChar(String s){
	char[] charArray = s.toCharArray();
	int value=0;
	for (int i=0;i<charArray.length;i++){
	    value=value*16+Character.digit(charArray[i],16);
	}
	try {
	    return new String(new byte[]{(byte)value},Streams.defaultEncoding).charAt(0);
	}
	catch (IOException e){
	    System.err.println("hexStringToChar: return-Umwandlung fehlgeschlagen");
	    return ' ';
	}
    }
   
    public void addTextNode(Element el, StringBuffer sb){
	if (sb.length()>0){
	    StringBuffer tempsb = new StringBuffer();
	    int c;
	    for (int i=0;i<sb.length();i++){
		c=(int)sb.charAt(i);
		if(c==0x9 || c==0xA || c==0xD || (0x20<=c && c<=0xD7FF)
		   || (0xE000<=c && c<=0xFFFD) || (0x10000<=c && c<=0x10FFFF))
		    tempsb.append((char)c);
	    }
	    Node te = el.getOwnerDocument().createTextNode(tempsb.toString());
	    el.appendChild(te);
	    sb.setLength(0);
	}    
    }

    public void parse(Element head, Element el, PushbackReader isr, int initC) throws IOException{
	int c;
	StringBuffer sb = new StringBuffer();
	while ((c=isr.read())!=-1){
	    if (c=='{'){
		addTextNode(el,sb);
		parse(head,el,isr,c);
		while ((c=isr.read())!=-1 && c!='}')
		    if(c=='\\') {
			for (int y=1;y<=3;y++)
			    isr.read();
			isr.unread("\\par {".toCharArray());
			return;
		    }
	    }
	    //TAGS
	    else if (c=='\\'){
		c=isr.read();
		// Umlaute und Sonderzeichen
		if(c=='\''){ 
		    StringBuffer localBuffer = new StringBuffer();
		    localBuffer.append((char)isr.read());
		    localBuffer.append((char)isr.read());
		    sb.append(hexStringToChar(localBuffer.toString()));
		}
		else if (!Character.isWhitespace((char)c)) {
		    StringBuffer tagBuffer = new StringBuffer();
		    tagBuffer.setLength(0);
		    tagBuffer.append((char)c);
		    while((c=isr.read())!=-1 && !Character.isWhitespace((char)c) 
			  && c!='\\' && c!='{' && c!='}'){
			tagBuffer.append((char)c);
		    }
		    isr.unread(c); 
		    String tag=tagBuffer.toString();
		    Format format=(Format) formats.get(tag);
		    if (format==null){
		    }
		    else if(format instanceof IgnoreFormat){
			IgnoreFormat igf = (IgnoreFormat) format;
			int bracketCount=0;
			// alles innerhalb diese Container ignorieren
			// falls vor diesem Tag ein {
			
			if(initC=='{'){
			    bracketCount=1;
			    while(bracketCount>0){
				c=isr.read();
				if (c=='{')
				    bracketCount++;
				else if(c=='}')
				    bracketCount--;
			    }
			    isr.unread(c);
			}
		    }
		    else if (format instanceof ParagraphFormat){
			ParagraphFormat pf = (ParagraphFormat) format;
			isr.unread(tag.toCharArray());
			isr.unread('\\');
			addTextNode(el,sb);
			return; 
		    }	
		    else if (format instanceof WhitespaceFormat){
		    }
		    else if (format instanceof EvaluateContainerFormat){
			EvaluateContainerFormat cf = (EvaluateContainerFormat) format;
		 	addTextNode(el,sb);
			parse(head,el,isr,c);
		    }
		    else if (format instanceof HeadContainerFormat){
			HeadContainerFormat cf = (HeadContainerFormat) format;
		 	addTextNode(el,sb);
	 		Element ne = el.getOwnerDocument().createElement(cf.tag);
 			head.appendChild(ne);
			parse(head,ne,isr,c);
		    }
		    else if (format instanceof ContainerFormat){
			ContainerFormat cf = (ContainerFormat) format;
			addTextNode(el,sb);
			Element ne = el.getOwnerDocument().createElement(cf.tag);
			el.appendChild(ne);
			parse(head,ne,isr,c);
		    }
		}
		else
		    System.err.println("!! Oops: Tag Anfang ohne eigentliches Tag !!");
	    }
	    else if (c=='}'){
		isr.unread(c);
		addTextNode(el,sb);
		return;
	    }
	    // Normaler Text
	    else {
		if(Character.isWhitespace((char)c))
		    sb.append(" ");
		else
		    sb.append((char)c);
	    }
	    initC=c; // ist der character, der vor einem eventuellen Tag steht
	} // while Schleife
	
	addTextNode(el,sb);
    }

    public Document translate(Reader reader) throws IOException{

	PushbackReader pr = new PushbackReader(reader,64);

	Document doc = DOMUtil.createDocument();
	Element root = doc.createElement(rootTag);
	Element head = doc.createElement(headTag);
	Element body = doc.createElement(bodyTag);
	doc.appendChild(root);
	root.appendChild(head);
	root.appendChild(body);

	ParagraphFormat pf=(ParagraphFormat) formats.get("par");
	int c;
	StringBuffer ub = new StringBuffer();
	while((c=pr.read())!=-1){
	    if (c=='\\'){  // schmeisst paragraph weg!!
		/*ub.setLength(0);
		  ub.append((char)c);
		  for(int y=1;y<=4;y++){
		  c=pr.read();
		  ub.append((char)c);
		  }  
		  String s = ub.toString();
		  if(s.equals("\\par"+" ") || s.equals("\\par"+"{") || s.equals("\\par"+"\\") ){
		  pr.unread(c); 
		  } 
		  else {
		  pr.unread(s.toCharArray()); 
		  }*/
		pr.unread('\\');
		pr.unread('\\');
		
	    }
	    else
		pr.unread(c);
	    Element par = doc.createElement(pf.tag);
	    body.appendChild(par);
	    parse(head,par,pr,0);
	}
	//	new DOMWriter (System.out).print(doc);
	return doc;
    }

    public static void main(String[] args) 
	throws IOException, SAXException, TransformerException {
	File cfgFile = new File("rtf2xml.xml");
	Document cfg = DOMUtil.parse(cfgFile, Streams.defaultEncoding);
	RtfToXml rtf = new RtfToXml(cfg.getDocumentElement());

	if (args.length==0){
	    System.err.println("kein File");
	    System.exit(0);
	}

	File file = new File(args[0]);   
	FileInputStream fis = new FileInputStream(file);
	Reader isr = Streams.streamToReader(fis);

	Document doc = rtf.translate(isr);

	new DOMWriter (System.out, Streams.defaultEncoding).print(doc);
	new DOMWriter (System.out, Streams.defaultEncoding).print(rtf.transform(doc,new File("rtf-html.xsl")));
	//System.err.println("Fertig");
    } 
}
